from dataclasses import dataclass
from logging import currentframe
from os import spawnlp
from typing import Any, Generic, List, Optional, Sequence, Tuple

import multiprocessing as mp
import itertools
import functools
import warnings
import instancelib
from numpy.linalg import LinAlgError
from numba.experimental import jitclass
from numba import jit, types, typed, int32, prange # type: ignore
from numba.typed import List as NumbaList

import pandas as pd

from ..activelearning.estimator import Estimator
import numpy as np
import numpy.typing as npt
from .rasch import EMRaschCombined

from ..typehints.typevars import KT, DT, VT, RT, LT

spec = [
    ("positive", int32[:]),
    ("estimate", int32[:]),
    ("negative", int32[:]),
    ("observed", int32[:]),
    ("positive_estimate", int32[:]),
    ("negative_estimate", int32[:]),
    ("positive_observed", int32[:]),
    ("negative_observed", int32[:]),
]

@jitclass(spec)
class Masks:
    def __init__(self, 
                 positive: npt.NDArray[Any], 
                 estimate: npt.NDArray[Any], 
                 negative: npt.NDArray[Any], 
                 observed: npt.NDArray[Any], 
                 positive_estimate: npt.NDArray[Any], 
                 negative_estimate: npt.NDArray[Any], 
                 positive_observed: npt.NDArray[Any], 
                 negative_observed: npt.NDArray[Any]):
        self.positive = positive
        self.estimate = estimate
        self.negative = negative
        self.observed = observed
        self.positive_estimate = positive_estimate
        self.negative_estimate = negative_estimate
        self.positive_observed = positive_observed
        self.negative_observed = negative_observed

def create_mask(poses: List[bool], estes: List[bool]) -> Masks:
    def convert(bool_list: Sequence[bool]) -> npt.NDArray[Any]:
        converted = np.array([i for (i,b) in enumerate(bool_list) if b], dtype=np.int32)
        return converted
    positive = convert(poses)
    estimate = convert(estes)
    neges = [not(m) for m in poses]
    obses = [not(m) for m in estes]
    negative = convert(neges)
    observed = convert(obses)
    positive_estimate = convert(NumbaList([est & pos for est, pos in zip(estes, poses)]))
    negative_estimate = convert(NumbaList([est & neg for est, neg in zip(estes, neges)]))
    positive_observed = convert(NumbaList([obs & pos for obs, pos in zip(obses, poses)]))
    negative_observed = convert(NumbaList([obs & neg for obs, neg in zip(obses, neges)]))
    return Masks(positive, estimate, 
                 negative, observed, 
                 positive_estimate, negative_estimate, 
                 positive_observed, negative_observed)
@jit(nopython=True) # type: ignore
def l2(b0: npt.NDArray[Any],
       design_mat: npt.NDArray[Any],
       counts: npt.NDArray[Any],
       lam: float = 0.0,
       tolerance: float = 1e-8,
       max_it: int = 10000,
       epsilon: float = 0.1
       ) -> npt.NDArray[Any]:
    counts = counts + epsilon
    current_tolerance = 1
    b = b0
    it = 0
    while current_tolerance > tolerance and it < max_it:
        it = it + 1

        mu = np.exp(design_mat @ b0)
        all_but_first_b0: npt.NDArray[Any] = b0[1:]
        lbpart: npt.NDArray[Any] = 2 * np.concatenate(
            (
                np.array([0.0]),
                (lam * all_but_first_b0)
        ))
        design_mat_t = design_mat.conj().transpose()
        db = design_mat_t @ counts - design_mat_t @ mu - lbpart
        repeat_count = b0.shape[0] - 1
        other_lb_part = np.hstack((np.array([0.0]), np.repeat(lam, repeat_count)))
        I = (design_mat_t @ np.diag(mu)) @ design_mat + \
            2 * np.diag(other_lb_part)
        try:
            invI = np.linalg.inv(I)
        except Exception: # LinAlgException
            # Change lambda if the matrix is not invertible
            lam = 1e-6 if lam == 0 else 10 * lam
        else:
            # If inversion succeeded, calculate new coefficients b
            b = b0 + invI @ db
            current_tolerance = np.sum(np.abs(b0 - b))
            b0 = b
    # if it >= max_it:
    #     warnings.warn(
    #         f"Exceeded maximum L2 iterations ({max_it}). Estimate might not be accurate")
    return b


@jit(nopython=True)  # type: ignore
def calc_deviance(y: npt.NDArray[Any], y_hat: npt.NDArray[Any]) -> float:
    pos_y_idx = np.where(y > 0)
    pos_y = y[pos_y_idx]
    deviance = (
        2 * np.sum(pos_y * np.log(pos_y / y_hat[pos_y_idx])) -
        2 * np.sum(y - y_hat))
    return deviance


def add_missing_count_rows(freq_df: pd.DataFrame, n_pos: float, n_neg: float) -> pd.DataFrame:
    pos_dict = {
        **{col: 0 for col in freq_df.columns},
        **{"count": n_pos, "positive": 1}
    }
    neg_dict = {
        **{col: 0 for col in freq_df.columns},
        **{"count": n_neg, "positive": 0}
    }
    new_freq_df = (
        freq_df
        .append(pos_dict, ignore_index=True)
        .append(neg_dict, ignore_index=True)
    )
    return new_freq_df


def l2_format(freq_df: pd.DataFrame, learner_cols: Sequence[str]) -> pd.DataFrame:
    order_cols = ["positive", *learner_cols]
    df = freq_df.sort_values(order_cols)
    df.insert(0, "intercept", 1)
    return df

@jit(nopython=True)  # type: ignore
def parameter_ci(estimate: float, se: float) -> Tuple[float, float]:
    var_min = estimate - 2 * se
    var_max = estimate + 2 * se
    return var_max, var_min

@jit(nopython=True)  # type: ignore
def calculate_estimate(intercept: float, pos: float) -> float:
    return np.exp(intercept + pos)

def calculate_ci_rasch(intercept: float,
                       positive: float,
                       intercept_se: float,
                       positive_se: float) -> Tuple[float, float]:
    intercepts = parameter_ci(intercept, intercept_se)
    positive_par = parameter_ci(positive, positive_se)
    combinations = itertools.product(intercepts, positive_par)
    results = list(itertools.starmap(calculate_estimate, combinations))
    min_result = min(results)
    max_result = max(results)
    return min_result, max_result

@jit(nopython=True)  # type: ignore
def initial_fit(n_not_read: float, design_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
    not_read = np.array([n_not_read])
    b0 = np.hstack(
        (
            np.log(not_read),
            np.repeat(0.0, design_mat.shape[1] - 1)
    ))
    return b0

@jit(nopython=True)  # type: ignore
def rasch_glm(design_mat: npt.NDArray[Any],
              efit: npt.NDArray[Any],
              counts: npt.NDArray[Any],
              b0: npt.NDArray[Any],
              masks: Masks
              ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    obs_counts = counts[masks.observed]
    beta = l2(b0, design_mat, counts)
    mfit = np.exp(design_mat @ beta)
    obs_fitted = mfit[masks.observed]
    deviance = calc_deviance(obs_counts, obs_fitted)
    return beta, mfit, deviance





@jit(nopython=True)  # type: ignore
def rasch_em(design_mat: npt.NDArray[Any], 
             counts: npt.NDArray[Any],
             n_dataset: float,
             masks: Masks,
             proportion: float,
             tolerance: float = 1e-5,
             max_it=1000) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    orig_counts = counts[masks.observed]
    n_not_read = n_dataset - np.sum(orig_counts)
    
    n_pos = proportion * n_not_read
    n_neg = (1 - proportion) * n_not_read
    efit = counts

    efit[masks.positive_estimate] = n_pos
    efit[masks.negative_estimate] = n_neg

    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = rasch_glm(design_mat, efit, counts, b0, masks)
    current_tolerance = 1
    # Expectation Maximization
    it = 0
    while current_tolerance > tolerance:
        old_deviance = deviance
        efit[masks.estimate] = mfit[masks.estimate] * n_not_read / np.sum(mfit[masks.estimate])
        beta, mfit, deviance = rasch_glm(design_mat, efit, counts, beta, masks)
        current_tolerance = old_deviance - deviance
        if it > max_it:
            #warnings.warn(
            #    f"Exceeded maximum EM iterations ({max_it}). Estimate might not be accurate")
            return beta, mfit, deviance
    return beta, mfit, deviance

@jit(nopython=True, parallel=True) # type: ignore
def rasch_numpy(design_mat: npt.NDArray[Any],
                counts: npt.NDArray[Any],
                n_dataset: float,
                masks: Masks,
                proportions: npt.NDArray[Any] = np.array([0.001, 0.01, 0.1, 0.5, 0.75, 0.9]),
                tolerance: float = 1e-5,
                max_it=1000
                ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    deviances = np.zeros(proportions.shape)
    betas = np.zeros((proportions.shape[0], design_mat.shape[1]))
    mfits = np.zeros((proportions.shape[0], design_mat.shape[0]))
    num_proportions = proportions.shape[0]
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    for idx in prange(num_proportions):
        proportion = proportions[idx]
        b0 = initial_fit(n_not_read, design_mat)
        beta, mfit, deviance = rasch_em(design_mat, counts, n_dataset, masks, proportion, tolerance, max_it)
        betas[idx,:] = beta
        mfits[idx,:] = mfit
        deviances[idx] = deviance
    best_beta = np.zeros((design_mat.shape[1],))
    best_mfit = np.zeros((design_mat.shape[0],))
    best_deviance = 2000000.0
    for idx in range(num_proportions):
        current_beta = betas[idx,:]
        current_mfit = mfits[idx,:]
        current_deviance = deviances[idx]
        if current_deviance < best_deviance:
            best_deviance = current_deviance
            best_beta = current_beta
            best_mfit = current_mfit
    return best_beta, best_mfit, best_deviance

@jit(nopython=True) # type: ignore
def rasch_numpy_single_thread(design_mat: npt.NDArray[Any],
                counts: npt.NDArray[Any],
                n_dataset: float,
                masks: Masks,
                proportions: npt.NDArray[Any] = np.array([0.001, 0.01, 0.1, 0.5, 0.75, 0.9]),
                tolerance: float = 1e-5,
                max_it=1000
                ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    deviances = np.zeros(proportions.shape)
    betas = np.zeros((proportions.shape[0], design_mat.shape[1]))
    mfits = np.zeros((proportions.shape[0], design_mat.shape[0]))
    num_proportions = proportions.shape[0]
    for idx in range(num_proportions):
        proportion = proportions[idx]
        beta, mfit, deviance = rasch_em(design_mat, counts, n_dataset, masks, proportion, tolerance, max_it)
        betas[idx,:] = beta
        mfits[idx,:] = mfit
        deviances[idx] = deviance
    best_beta = np.zeros((design_mat.shape[1],))
    best_mfit = np.zeros((design_mat.shape[0],))
    best_deviance = 2000000.0
    for idx in range(num_proportions):
        current_beta = betas[idx,:]
        current_mfit = mfits[idx,:]
        current_deviance = deviances[idx]
        if current_deviance < best_deviance:
            best_deviance = current_deviance
            best_beta = current_beta
            best_mfit = current_mfit
    return best_beta, best_mfit, best_deviance


def rasch_estimate_parametric(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 50) -> Tuple[float, float, float]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore

    beta, mfit, deviance = rasch_numpy(
        design_mat, counts, n_dataset, masks, tolerance=tolerance, max_it=max_it)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    p_vals = mfit / np.sum(mfit)
    multinomial_fits: npt.NDArray[Any] = np.random.multinomial(
        n_dataset, p_vals, multinomial_size)
    low_estimate = positive_estimate
    high_estimate = positive_estimate
    if positive_estimate <= 2:
        workload = [
            (design_mat, np.array(sample), n_dataset, masks) 
            for sample in multinomial_fits.tolist()]
        with mp.Pool(7) as pool:
            results = pool.starmap(rasch_numpy, workload)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                for (_, fitted, _) in results]
        low_estimate = np.percentile(estimates, 2.5)
        high_estimate = np.percentile(estimates, 97.5)
    return positive_estimate, low_estimate, high_estimate

@jit(nopython=True, parallel=True) # type: ignore
def rasch_parallel(design_mat: npt.NDArray[Any], 
                   rounded_mfits: npt.NDArray[Any], 
                   n_dataset: int, masks: Masks):
    deviances = np.zeros(rounded_mfits.shape[0])
    betas = np.zeros((rounded_mfits.shape[0], design_mat.shape[1]))
    mfits = np.zeros((rounded_mfits.shape[0], design_mat.shape[0]))
    num_mfits = rounded_mfits.shape[0]
    for idx in prange(num_mfits):
        mfit_sample = rounded_mfits[idx,:]
        beta, mfit, deviance = rasch_numpy_single_thread(design_mat, mfit_sample, n_dataset, masks) 
        betas[idx,:] = beta
        mfits[idx,:] = mfit
        deviances[idx] = deviance
    results = NumbaList()
    for idx in range(num_mfits):
        result = (betas[idx,:], mfits[idx,:], deviances[idx])
        results.append(result)
    return results
        
def rasch_estimate_parametric_approx(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 50) -> Tuple[float, float, float]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore

       
    beta, mfit, deviance = rasch_numpy(
        design_mat, counts, n_dataset, masks, tolerance=tolerance, max_it=max_it)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    
    # Calculate standard error on predictors
    mat_w = np.diag(mfit)
    design_mat_t = design_mat.conj().transpose()
    vcov = np.linalg.inv(design_mat_t @ mat_w @ design_mat)
    
    predictors_sampled: npt.NDArray[Any] = np.random.multivariate_normal(beta, vcov, size=50)
    sampled_mfits = np.exp(design_mat @ predictors_sampled.T).T
    rounded_mfits = np.round(sampled_mfits)
    low_estimate = positive_estimate
    high_estimate = positive_estimate
    obs_pos = np.sum(counts[masks.positive_observed])
    if True:
        results = rasch_parallel(design_mat, rounded_mfits, n_dataset, masks)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                for (_, fitted, _) in results]
        low_estimate = np.percentile(estimates, 2.5)
        high_estimate = np.percentile(estimates, 97.5)
    return positive_estimate, low_estimate, high_estimate




class EMRaschRidgeParametricPython(
        EMRaschCombined[KT, DT, VT, RT, LT],
        Generic[KT, DT, VT, RT, LT]):
    def __init__(self):
        super().__init__()
        self.df: Optional[pd.DataFrame] = None
        self.est = float("nan")
        self.est_low = float("nan")
        self.est_high = float("nan")

    def _start_r(self) -> None:
        pass

    def calculate_estimate(self,
                           estimator: Estimator[Any, KT, DT, VT, RT, LT],
                           label: LT) -> Tuple[float, float, float]:
        pos_count = estimator.env.labels.document_count(label)
        dataset_size = len(estimator.env.dataset)
        df = self.get_occasion_history(estimator, label)
        if self.df is None or not self.df.equals(df):
            self.df = df
            self.est, self.est_low, self.est_up = rasch_estimate_parametric_approx(
                df, dataset_size)
        horizon = self.est + pos_count
        horizon_low = self.est_low + pos_count
        horizon_up = self.est_up + pos_count
        return horizon, horizon_low, horizon_up


from .catalog import EnvironmentCatalog as ENV
from .memory import DataPointEnvironment

from ..factory import AbstractBuilder, ObjectFactory
from ..module.component import Component


class EnvironmentBuilder(AbstractBuilder):
    def __call__(self, environment_type: ENV.Type, **kwargs): # type: ignore
        return self._factory.create(environment_type, **kwargs)


class EnvironmentFactory(ObjectFactory):
    def __init__(self) -> None:
        super().__init__()
        self.register_builder(Component.ENVIRONMENT, EnvironmentBuilder())
        self.register_constructor(ENV.Type.MEMORY, DataPointEnvironment)


from __future__ import annotations

from typing import Generic, Sequence, TypeVar, Iterable, Dict, Any, Set
from joblib import Memory

import numpy as np
import numpy.typing as npt

from instancelib.instances.memory import MemoryBucketProvider
from instancelib.instances.hdf5 import HDF5Instance, HDF5Provider
from instancelib.labels.memory import MemoryLabelProvider
from ..history import MemoryLogger

from .base import AbstractEnvironment

# TODO Adjust MemoryEnvironment Generic Type (ADD ST)

class HDF5Environment(AbstractEnvironment[int, str, npt.NDArray[Any], str, str]):
    def __init__(
            self,
            dataset: HDF5Provider,
            unlabeled: HDF5Provider,
            labeled: HDF5Provider,
            labelprovider: MemoryLabelProvider[int, str],
            logger: MemoryLogger[int, str, Any],
            truth: MemoryLabelProvider[int, str]
        ):
        self._dataset = dataset
        self._unlabeled = unlabeled
        self._labeled = labeled
        self._labelprovider = labelprovider
        self._named_providers: Dict[str, HDF5Provider] = dict()
        self._logger = logger
        self._storage: Dict[str, Any] = dict()
        self._truth = truth

    @classmethod
    def from_data(cls, 
            target_labels: Iterable[str], 
            indices: Sequence[int], 
            data: Sequence[str], 
            ground_truth: Sequence[Set[str]],
            data_location: str,
            vector_location: str) -> HDF5Environment:
        dataset = HDF5Provider.from_data_and_indices(indices, data, data_location, vector_location)
        unlabeled = MemoryBucketProvider(dataset, dataset.key_list)
        labeled = MemoryBucketProvider(dataset, [])
        labels = MemoryLabelProvider[int, str].from_data(target_labels, indices, [])
        logger = MemoryLogger[int, str, Any](labels)
        truth = MemoryLabelProvider[int, str].from_data(target_labels, indices, ground_truth)
        return cls(dataset, unlabeled, labeled, labels, logger, truth)

    @classmethod
    def from_environment(cls, 
                         environment: AbstractEnvironment[int, str, npt.NDArray[Any], str, str], 
                         data_location: str = "", vector_location: str = "", 
                         shared_labels: bool = True, *args, **kwargs) -> HDF5Environment:
        if isinstance(environment.dataset, HDF5Provider):
            dataset = environment.dataset
        else:
            dataset = HDF5Provider.from_provider(environment.dataset, data_location, vector_location)
        unlabeled = MemoryBucketProvider(dataset, environment.unlabeled.key_list)
        labeled = MemoryBucketProvider(dataset, environment.labeled.key_list)
        if isinstance(environment.labels, MemoryLabelProvider) and shared_labels:
            labels: MemoryLabelProvider[int, str] = environment.labels
        else:
            labels = MemoryLabelProvider[int, str](environment.labels.labelset, {}, {}) # type: ignore
        if isinstance(environment.logger, MemoryLogger):
            logger: MemoryLogger[int, str, Any] = environment.logger
        else:
            logger = MemoryLogger[int, str, Any](labels)
        if isinstance(environment.truth, MemoryLabelProvider):                
            truth = environment.truth
        else:
            truth = MemoryLabelProvider[int, str](labels.labelset, {}, {})
        return cls(dataset, unlabeled, labeled, labels, logger, truth)

    @classmethod
    def from_environment_only_data(cls, 
                                   environment: AbstractEnvironment[int, str, npt.NDArray[Any], str, str],
                                   data_location: str, vector_location: str) -> HDF5Environment:
        if isinstance(environment.dataset, HDF5Provider):
            dataset = environment.dataset
        else:
            dataset = HDF5Provider.from_provider(environment.dataset, data_location, vector_location)
        unlabeled = MemoryBucketProvider(dataset, environment.dataset.key_list)
        labeled = MemoryBucketProvider(dataset, [])
        labels = MemoryLabelProvider[int, str](environment.labels.labelset, {}, {}) # type: ignore
        if isinstance(environment.logger, MemoryLogger):
            logger: MemoryLogger[int, str, Any] = environment.logger
        else:
            logger = MemoryLogger[int, str, Any](labels)
        if isinstance(environment.truth, MemoryLabelProvider):                
            truth = environment.truth
        else:
            truth = MemoryLabelProvider[int, str](labels.labelset, {}, {})
        return cls(dataset, unlabeled, labeled, labels, logger, truth)

    def create_named_provider(self, name: str) -> HDF5Provider:
        self._named_providers[name] = MemoryBucketProvider(self._dataset, [])
        return self._named_providers[name]

    def get_named_provider(self, name: str) -> HDF5Provider:
        if name in self._named_providers:
            self.create_named_provider(name)
        return self._named_providers[name]

    def create_empty_provider(self) -> MemoryBucketProvider:
        return MemoryBucketProvider(self._dataset, [])

    @property
    def dataset(self):
        return self._dataset

    @property
    def unlabeled(self):
        return self._unlabeled

    @property
    def labeled(self):
        return self._labeled

    @property
    def labels(self):
        return self._labelprovider

    @property
    def truth(self):
        return self._truth

    @property
    def logger(self): # TODO Replace Any Type
        return self._logger

    def store(self, key: str, value: Any) -> None:
        self._storage[key] = value

    
    def storage_exists(self, key: str) -> bool:
        return key in self._storage

    def retrieve(self, key: str) -> Any:
        return self._storage[key]

    
    
    



        




# from typing import Any, Optional, Set, Tuple

# import numpy as np
# import numpy.typing as npt

# from .base import AbstractActiveLearner
# from .oracles import OracleFunction

# from ..instances import Instance

# class LeastConfidence(AbstractActiveLearner):
#     def sample(self) -> Optional[Instance]:
#         """Select the document whose  posterior probability is most near 0.5

#         Returns
#         -------
#         Optional[Instance]
#             A tuple containing:
#                 - The document identifier
#                 - The document vector
#                 - The document data
#         """
#         def least_confidence(vector: npt.NDArray[Any]) -> float:
#             prob_vec = self.estimator.predict_proba(vector)
#             min_prob = np.amin(np.abs(prob_vec - 0.5))
#             return min_prob
#         doc_id = self.argmin(self.unlabeled, least_confidence)
#         if doc_id is not None:
#             return self.unlabeled[doc_id]
#         return None

#     def initialize(self, f_oracle: OracleFunction, it_max: int) -> None:
#         """Label `it_max` randomly selected documents 

#         Parameters
#         ----------
#         f_oracle :
#             Oracle function 
#         it_max : int
#             The number of iterations / documents
#         """
#         for _ in range(it_max):
#             doc_id, vec, data = super().sample()
#             label = f_oracle(doc_id, vec, data, self.labels)
#             self.label_instance(doc_id, label)

#     def iterate_al(self, f_oracle: OracleFunction) -> None:
#         """Perform one iteration of the active learning cycle

#         Parameters
#         ----------
#         f_oracle : [type]
#             Oracle function
#         """
#         # Select cluster and sample data
#         instance = self.sample()

#         # Query the oracle and label the instance
#         label = f_oracle(instance, self.labels)
#         self.label_instance(instance.identifier, label)

#         # Retrain the Machine learning model
#         self.fit()


# class EntropySampling(LeastConfidence):
#     def sample(self) -> Optional[Instance]:
#         """Select the document whose  posterior probability is most near 0.5

#         Returns
#         -------
#         Optional[Instance]
#             A tuple containing:
#                 - The document identifier
#                 - The document vector
#                 - The document data
#         """
#         def entropy(vector: npt.NDArray[Any]) -> float:
#             prob_vec = self.estimator.predict_proba(vector)
#             entropy = -np.matmul(prob_vec, np.log(prob_vec))
#             return entropy
#         doc_id = self.argmax(self.unlabeled, entropy)
#         if doc_id is not None:
#             return self.unlabeled[doc_id]
#         return None



import itertools
from typing import FrozenSet, Generic, Iterable

from instancelib.machinelearning.base import AbstractClassifier

from ..utils.func import union

from ..environment.base import AbstractEnvironment

from ..typehints.typevars import IT
from instancelib.pertubations.base import ProviderPertubator
from instancelib.typehints.typevars import DT, KT, LT, RT, VT, LMT, PMT


class SyntheticTrainMixIn(Generic[IT, KT, DT, VT, RT, LT, LMT, PMT]):
    
    env: AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    synthesizer: ProviderPertubator[IT, KT, DT, VT, RT]
    synthetic_labels: FrozenSet[LT]
    classifier: AbstractClassifier[IT, KT, DT, VT, RT, LT, LMT, PMT]
    
    def __init__(self, synthesizer: ProviderPertubator[IT, KT, DT, VT, RT], synthetic_labels: Iterable[LT]) -> None:
        self.synthesizer = synthesizer
        self.synthetic_labels = frozenset(synthetic_labels)
    
    def retrain(self) -> None:
        """Retrain the classifier using the information in the labeled set
        and generated instances by the synthesizer

        Raises
        ------
        NotInitializedException
            If the AL method has no attached Environment
        """
        minority_ids = union(*(self.env.labels.get_instances_by_label(label) for label in self.synthetic_labels))
        minority_labeled = minority_ids.intersection(self.env.labeled)
        minority_provider = self.env.create_bucket(minority_labeled)
        synthetic_data = self.synthesizer(minority_provider)
        train_provider = self.env.combine(self.env.labeled, synthetic_data)
        self.classifier.fit_provider(train_provider, self.env.labels)




from .base import ActiveLearner
from typing import Any, Callable, Iterable, List, Optional, TypeVar

import numpy as np # type: ignore

from instancelib.instances import Instance

from ..typehints import KT, DT, LT, RT, VT, IT

OracleFunction = Callable[[Instance, Iterable[LT]], LT]

def console_text_oracle(
        doc: Instance,
        labels: Iterable[LT]) -> List[LT]:
    label_dict = dict(enumerate(labels, start=1))
    qstr =  "Please label the following instance: \n"
    qstr += "==================================== \n"
    qstr += "{} \n".format(doc.representation) 
    qstr += "==================================== \n"
    qstr += "Document ID: {} \n".format(doc.identifier)
    qstr += "Vector: {} \n".format(doc.vector)
    qstr += "==================================== \n"
    for i, label in label_dict.items():
        qstr += "{} => {}\n".format(i, label)
    chosen_label = int(input(qstr))
    return [label_dict[chosen_label]]

def label_loop(al: ActiveLearner[IT, KT, DT, VT, RT, LT]) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
    qstr = "Press ENTER to continue or one of the following options"
    qstr += "q => quit\n"
    qstr += "u => update ordering"
    continue_choice = input(qstr)
    if  continue_choice == "q":
        return al
    elif continue_choice == "u":
        print("Retraining the classifier...")
        al.update_ordering()
    print("Retrieving the next document... \n")
    
    labelset = al.env.labels.labelset
    label_dict = dict(enumerate(labelset, start=1))
    
    doc = next(al)
    
    lstr =  "Please label the following instance: \n"
    lstr += "==================================== \n"
    lstr += f"{doc.representation} \n"
    lstr += "==================================== \n"
    lstr += f"Document ID: {doc.identifier} \n"
    lstr += f"Vector: {doc.vector} \n"
    lstr += "==================================== \n"
    for i, label in label_dict.items():
        lstr += f"{i} => {label}\n"
    # Request labels from user 
    chosen_labels = [label_dict[int(input(lstr))]]
    
    # Update state
    al.env.labels.set_labels(doc, *chosen_labels)
    al.set_as_labeled(doc)
    return label_loop(al)
    
        




from .base import BaseBalancer, IdentityBalancer
from .catalog import BalancerCatalog
from .double import DoubleBalancer
from .factory import BalancerFactory
from .randomoversampling import RandomOverSampler
from .undersample import UndersampleBalancer



from __future__ import annotations

from typing import Generic, Iterable, TypeVar, Any
from abc import ABC, abstractmethod, abstractclassmethod
from ..history import BaseLogger
from instancelib import InstanceProvider, Instance
from instancelib.typehints import KT, DT, VT, RT, LT
from instancelib.labels import LabelProvider

import instancelib as ins

IT = TypeVar("IT", bound="Instance[Any, Any, Any, Any]", covariant=True)



class AbstractEnvironment(ins.AbstractEnvironment[IT, KT, DT, VT, RT, LT], 
                          ABC, Generic[IT, KT, DT, VT, RT, LT]):

    @property
    @abstractmethod
    def unlabeled(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        """This `InstanceProvider` contains all unlabeled instances.
        `ActiveLearner` methods sample instances from this provider/

        Returns
        -------
        InstanceProvider[KT, DT, VT, RT]
            An `InstanceProvider` that contains all unlabeld instances
        """        
        raise NotImplementedError

    @property
    @abstractmethod
    def labeled(self) -> InstanceProvider[IT, KT, DT, VT, RT]:
        """This `InstanceProvider` contains all labeled instances.
        `ActiveLearner` may use this provider to train a classifier
        to sample instances from the `unlabeled` provider.

        Returns
        -------
        InstanceProvider[KT, DT, VT, RT]
            An `InstanceProvider` that contains all labeled instances
        """        
        raise NotImplementedError

    @property
    @abstractmethod
    def logger(self) -> BaseLogger[KT, LT, Any]:
        """This property contains an implementation of a `BaseLogger`.
        The logger can log events like sampling an instance, or its labeling.

        Returns
        -------
        BaseLogger[KT, LT, Any]
            The logger
        """        
        raise NotImplementedError

    @property
    @abstractmethod
    def truth(self) -> LabelProvider[KT, LT]:
        """This property contains a `LabelProvider` that maps 
        instances to *ground truth* labels and vice-versa. 
        This can be used for simulation purposes if you want
        to assess the performance of an AL algorithm on a dataset
        with a ground truth.

        Returns
        -------
        LabelProvider[KT, LT]
            The label provider that contains the ground truth labels
        """
        raise NotImplementedError

    @abstractclassmethod
    def from_environment(cls, 
                         environment: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
                         *args, **kwargs
                        ) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        """Create a new independent environment with the same state.
        Implementations may enable conversion from and to several types
        of Enviroments.

        Parameters
        ----------
        environment : AbstractEnvironment[KT, DT, VT, RT, LT]
            The environment that should be duplicated

        Returns
        -------
        AbstractEnvironment[KT, DT, VT, RT, LT]
            A new independent with the same state
        """        
        raise NotImplementedError

    


from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer # type: ignore

from ..factory import AbstractBuilder, ObjectFactory
from ..module.component import Component

from .base import SeparateContextVectorizer, StackVectorizer
from .catalog import FECatalog as FE
#from .contextinstance import ContextVectorizer
from .doc2vec import Doc2VecVectorizer
from .textinstance import TextInstanceVectorizer
from .textsklearn import SklearnVectorizer


class FeatureExtractorBuilder(AbstractBuilder):
    def __call__(self, datatype, **kwargs):
        return self._factory.create(datatype, **kwargs)

class VectorizerBuilder(AbstractBuilder):
    def __call__(self, vec_type, **kwargs):
        return self._factory.create(vec_type, **kwargs)

class TextInstanceVectorizerBuilder(AbstractBuilder):
    def __call__(self, **kwargs):
        vectorizer = self._factory.create(Component.VECTORIZER, **kwargs)
        return TextInstanceVectorizer(vectorizer)

# class ContextInstanceVectorizerBuilder(AbstractBuilder):
#     def __call__(self, **kwargs):
#         vectorizer = self._factory.create(Component.VECTORIZER, **kwargs)
#         return ContextVectorizer(vectorizer)

class StackVectorizerBuilder(AbstractBuilder):
    def __call__(self, vectorizers, **kwargs):
        vectorizers_list = [
            self._factory.create(Component.VECTORIZER, **config)
            for config in vectorizers]
        return StackVectorizer(*vectorizers_list)

class SeparateContextVectorizerBuider(AbstractBuilder):
    def __call__(self, vectorizer_data, vectorizer_context, **kwargs):
        data_vectorizer = self._factory.create(Component.VECTORIZER, **vectorizer_data)
        context_vectorizer = self._factory.create(Component.VECTORIZER, **vectorizer_context)
        return SeparateContextVectorizer(data_vectorizer, context_vectorizer)

class SklearnVectorizerBuilder(AbstractBuilder):
    def __call__(self, sklearn_vec_type, sklearn_config, storage_location = None, **_ignoredkwargs):
        sklearn_vec = self._factory.create(sklearn_vec_type, **sklearn_config)
        return SklearnVectorizer(sklearn_vec, storage_location)

class TfIDFVectorizerBuilder(AbstractBuilder):
    def __call__(self, **configuration) -> TfidfVectorizer:
        return TfidfVectorizer(**configuration)

class CountVectorizerBuilder(AbstractBuilder):
    def __call__(self, **configuration) -> CountVectorizer:
        return CountVectorizer(**configuration)


class FeatureExtractionFactory(ObjectFactory):
    def __init__(self) -> None:
        super().__init__()
        self.register_builder(Component.FEATURE_EXTRACTION, FeatureExtractorBuilder())
        self.register_builder(Component.VECTORIZER, VectorizerBuilder())
        # Data extractors
        # self.register_builder(FE.DataType.CHATMESSAGES, 
        #                  ContextInstanceVectorizerBuilder())
        self.register_builder(FE.DataType.TEXTINSTANCE,
                         TextInstanceVectorizerBuilder())
        # Vectorizer Containers
        self.register_builder(FE.VectorizerType.STACK, StackVectorizerBuilder())
        self.register_builder(FE.VectorizerType.SKLEARN, SklearnVectorizerBuilder())
        self.register_builder(FE.VectorizerType.DUALVEC, SeparateContextVectorizerBuider())

        # Actual vectorizers
        self.register_constructor(FE.VectorizerType.DOC2VEC, Doc2VecVectorizer)
        self.register_constructor(FE.SklearnVecType.TFIDF_VECTORIZER, TfidfVectorizer)
        self.register_constructor(FE.SklearnVecType.COUNT_VECTORIZER, CountVectorizer)



from instancelib.utils.func import all_subsets
import multiprocessing as mp
from typing import Any, Dict, FrozenSet, Generic, List, Optional, Sequence, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd

from ..activelearning.estimator import Estimator
from ..typehints import DT, KT, LT, RT, VT
from .rasch import NonParametricRasch
from .rasch_python import calc_deviance, l2


def l2_format(freq_df: pd.DataFrame, learner_cols: Sequence[str]) -> pd.DataFrame:
    df = freq_df.sort_values(learner_cols)
    df.insert(0, "intercept", 1)
    return df

def glm(design_mat: npt.NDArray[Any], counts: npt.NDArray[Any]) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    b0 = np.hstack(
        [
            np.repeat(1, design_mat.shape[1])
        ])
    beta = l2(b0, design_mat, counts, 0)
    mfit = np.exp(design_mat @ beta)
    deviance = calc_deviance(counts, mfit)
    return beta, mfit, deviance

def rasch_estimate(freq_df: pd.DataFrame, 
                   n_dataset: int,
                   proportion: float = 0.1,
                   tolerance: float = 1e-5, 
                   max_it: int = 2000) -> Tuple[float, float, float]:
    


    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(freq_df.filter(regex=r"^learner")
        )
    df_formatted = l2_format(freq_df, learner_cols)
   
    design_mat = (
        df_formatted.loc[:, df_formatted.columns != 'count'] # type: ignore
            .values) # type: ignore
    
    obs_counts: npt.NDArray[Any] = df_formatted["count"].values # type: ignore
    total_found = np.sum(obs_counts)
    
    beta, mfit, deviance = glm(design_mat, obs_counts)
    estimate =np.exp(beta[0])
    horizon_estimate = total_found + estimate

    fitted: npt.NDArray[Any] = np.concatenate((np.array([estimate]), mfit))
    p_vals = fitted / np.sum(fitted)

    if not np.isnan(estimate):
        multinomial_fits: npt.NDArray[Any] = np.random.multinomial(horizon_estimate, p_vals, max_it)
        only_observable_counts: List[List[float]] = multinomial_fits[:,1:].tolist()
        
        workload = [(design_mat, np.array(m_count)) for m_count in only_observable_counts]
        with mp.Pool(7) as pool:
            results = pool.starmap(glm, workload)
        
        estimates = [np.exp(result[0][0]) for result in results]
        
        middle_estimate = np.percentile(estimates, 50)
        low_estimate = np.percentile(estimates, 2.5)
        high_estimate = np.percentile(estimates, 97.5)

        lower_bound = low_estimate
        upper_bound = high_estimate
    
        return middle_estimate, lower_bound, upper_bound
    return horizon_estimate, horizon_estimate, horizon_estimate


class ParametricRaschPython(NonParametricRasch[KT, DT, VT, RT, LT], 
                      Generic[KT, DT, VT, RT, LT]):
    name = "RaschParametric"
    def __init__(self):
        super().__init__()
        self.df: Optional[pd.DataFrame] = None
        self.est = float("nan")
        self.est_low = float("nan")
        self.est_high = float("nan")
    
    def _start_r(self) -> None:
        pass
          
    def calculate_estimate(self, 
                           estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                           label: LT) -> Tuple[float, float, float]:
        pos_count = estimator.env.labels.document_count(label)
        dataset_size = len(estimator.env.dataset)
        df = self.get_occasion_history(estimator, label)
        if self.df is None or not self.df.equals(df):
            self.df = df        
            self.est, self.est_low, self.est_up = rasch_estimate(df, dataset_size)
        horizon = self.est + pos_count
        horizon_low = self.est_low + pos_count
        horizon_up = self.est_up + pos_count
        return horizon, horizon_low, horizon_up

class ParametricRaschHMore(ParametricRaschPython[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    
    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        interaction_cols = {
            f"h{i-1}": len(all_subsets(learner_set, i, i)) 
            for i in range(2, len(all_learners) + 1)
        }
        final_row = {
            **learner_cols,
            **interaction_cols,
            **count_col
        }
        return final_row

class ParametricRaschSingle(ParametricRaschPython[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    
    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        interaction_col = {"h1": len(
            all_subsets(
                learner_set, 
                2, len(all_learners) - 1)) 
        }
        final_row = {
            **learner_cols,
            **interaction_col,
            **count_col
        }
        return final_row


from __future__ import annotations

import logging
import pickle
from typing import (Any, FrozenSet, Iterable, List, Sequence, Set, Tuple,
                    TypeVar)

import numpy as np
import numpy.typing as npt
from instancelib import Instance

from sklearn.base import ClassifierMixin, TransformerMixin  # type: ignore

from ..balancing import BaseBalancer, IdentityBalancer
from ..environment import AbstractEnvironment
from ..utils import SaveableInnerModel
from ..utils.func import list_unzip
from .base import AbstractClassifier

LOGGER = logging.getLogger(__name__)

class SkLearnClassifier(SaveableInnerModel, AbstractClassifier[int, npt.NDArray[Any], str, npt.NDArray[Any], npt.NDArray[Any]]):
    _name = "Sklearn"

    def __init__(
            self,
            estimator: ClassifierMixin, encoder: TransformerMixin, balancer: BaseBalancer = IdentityBalancer(),
            storage_location=None, filename=None) -> None:
        SaveableInnerModel.__init__(self, estimator, storage_location, filename)
        self.encoder = encoder 
        self._fitted = False
        self._target_labels: FrozenSet[str] = frozenset()
        self.balancer = balancer

    def __call__(self, environment: AbstractEnvironment[Any, int, Any, npt.NDArray[Any], Any, str]) -> SkLearnClassifier:
        self._target_labels = frozenset(environment.labels.labelset)
        self.encoder.fit(list(self._target_labels)) # type: ignore
        return self

    def encode_labels(self, labels: Iterable[str]) -> npt.NDArray[Any]:
        return self.encoder.transform(list(set(labels))) # type: ignore

    def decode_vector(self, vector: npt.NDArray[Any]) -> Sequence[FrozenSet[str]]:
        labelings = self.encoder.inverse_transform(vector).tolist() # type: ignore
        return [frozenset([labeling]) for labeling in labelings]

    def get_label_column_index(self, label: str) -> int:
        label_list = self.encoder.classes_.tolist() # type: ignore
        return label_list.index(label)

    @SaveableInnerModel.load_model_fallback
    def fit(self, x_data: npt.NDArray[Any], y_data: npt.NDArray[Any]):
        assert x_data.shape[0] == y_data.shape[0]
        x_resampled, y_resampled = self.balancer.resample(x_data, y_data)
        LOGGER.info("[%s] Balanced / Resampled the data", self.name)
        self.innermodel.fit(x_resampled, y_resampled) # type: ignore
        LOGGER.info("[%s] Fitted the model", self.name)
        self._fitted = True

    def encode_xy(self, instances: Sequence[Instance[int, Any, npt.NDArray[Any], Any]], labelings: Sequence[Iterable[str]]):
        def yield_xy():
            for ins, lbl in zip(instances, labelings):
                if ins.vector is not None:
                    yield ins.vector, self.encode_labels(lbl)
        x_data, y_data = list_unzip(yield_xy())
        x_fm = np.vstack(x_data)
        y_lm = np.vstack(y_data)
        if y_lm.shape[1] == 1:
            y_lm = np.reshape(y_lm, (y_lm.shape[0],))
        return x_fm, y_lm

    def encode_x(self, instances: Sequence[Instance[int, Any, npt.NDArray[Any], Any]]) -> npt.NDArray[Any]:
        # TODO Maybe convert to staticmethod
        x_data = [
            instance.vector for instance in instances if instance.vector is not None]
        x_vec = np.vstack(x_data)
        return x_vec

    def encode_y(self, labelings: Sequence[Iterable[str]]) -> npt.NDArray[Any]:
        y_data = [self.encode_labels(labeling) for labeling in labelings]
        y_vec = np.vstack(y_data)
        if y_vec.shape[1] == 1:
            y_vec = np.reshape(y_vec, (y_vec.shape[0],))
        return y_vec

    @SaveableInnerModel.load_model_fallback
    def predict_proba(self, x_data: npt.NDArray[Any]) -> npt.NDArray[Any]:
        assert self.innermodel is not None
        return self.innermodel.predict_proba(x_data) 

    @SaveableInnerModel.load_model_fallback
    def predict(self, x_data: npt.NDArray[Any]) -> npt.NDArray[Any]:
        assert self.innermodel is not None
        return self.innermodel.predict(x_data)

    def predict_instances(self, instances: Sequence[Instance[int, Any, npt.NDArray[Any], Any]]) -> Sequence[FrozenSet[str]]:
        x_vec = self.encode_x(instances)
        y_pred = self.predict(x_vec)
        return self.decode_vector(y_pred)

    def predict_proba_instances(self, instances: Sequence[Instance[int, Any, npt.NDArray[Any], Any]]) -> Sequence[FrozenSet[Tuple[str, float]]]:
        x_vec = self.encode_x(instances)
        y_pred = self.predict_proba(x_vec).tolist()
        label_list: List[str] = self.encoder.classes_.tolist() # type: ignore
        y_labels: List[FrozenSet[Tuple[str, float]]] = [
            frozenset(zip(label_list, y_vec)) # type: ignore
            for y_vec in y_pred
        ]
        return y_labels
    
    @property
    def name(self) -> str:
        return f"{self._name} :: {self.innermodel.__class__}"
        
    def fit_vectors(self, x_data: Sequence[npt.NDArray[Any]], labels: Sequence[FrozenSet[str]]):
        x_mat = np.vstack(x_data)
        y_vec = self.encode_y(labels)
        self.fit(x_mat, y_vec)

    def fit_instances(self, instances: Sequence[Instance[int, Any, npt.NDArray[Any], Any]], labels: Sequence[FrozenSet[str]]):
        assert len(instances) == len(labels)
        x_train_vec, y_train_vec = self.encode_xy(instances, labels)
        self.fit(x_train_vec, y_train_vec)

    @property
    def fitted(self) -> bool:
        return self._fitted




class MultilabelSkLearnClassifier(SkLearnClassifier):
    _name = "Multilabel Sklearn"
    def __call__(self, environment: AbstractEnvironment[Any, int, Any, npt.NDArray[Any], Any, str]) -> SkLearnClassifier:
        self._target_labels = frozenset(environment.labels.labelset)
        self.encoder.fit(list(map(lambda x: {x}, self._target_labels))) # type: ignore
        return self

    def encode_labels(self, labels: Iterable[str]) -> npt.NDArray[Any]:
        return self.encoder.transform([list(set(labels))]) # type: ignore

    def decode_vector(self, vector: npt.NDArray[Any]) -> Sequence[FrozenSet[str]]:
        labelings = self.encoder.inverse_transform(vector) # type: ignore
        return [frozenset(labeling) for labeling in labelings]



import functools
from abc import ABC, abstractmethod
from os import PathLike
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import matplotlib.pyplot as plt  # type: ignore
import numpy as np  # type: ignore
import pandas as pd
from instancelib.analysis.base import classifier_performance
from instancelib.instances.base import InstanceProvider
from instancelib.labels.base import LabelProvider
from instancelib.typehints.typevars import KT

from ..activelearning import ActiveLearner
from ..activelearning.ensembles import AbstractEnsemble
from ..activelearning.insclass import ILMLBased
from ..estimation.base import AbstractEstimator
from ..utils.func import flatten_dicts
from .analysis import process_performance
from .experiments import ExperimentIterator  # type: ignore

LT = TypeVar("LT")


def name_formatter(learner: ActiveLearner[Any, Any, Any, Any, Any, Any]) -> str:
    name, label = learner.name
    if label is not None:
        return f"{name}_{label}"
    return name


class AbstractPlotter(ABC, Generic[LT]):
    @abstractmethod
    def update(self, activelearner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        raise NotImplementedError

    @abstractmethod
    def show(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError


class ExperimentPlotter(ABC, Generic[LT]):
    @abstractmethod
    def update(
        self,
        exp_iterator: ExperimentIterator[Any, Any, Any, Any, Any, LT],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def show(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError

    @abstractmethod
    def print_last_stats(self) -> None:
        raise NotImplementedError


class ClassificationPlotter(AbstractPlotter[LT], Generic[KT, LT]):
    def __init__(
        self,
        test_set: InstanceProvider[Any, KT, Any, Any, Any],
        ground_truth: LabelProvider[KT, LT],
    ):
        self.test_set = list(test_set.values())
        self.ground_truth = ground_truth
        self.result_frame = pd.DataFrame()
        self.labelset = ground_truth.labelset

    def update(self, activelearner: ActiveLearner[Any, KT, Any, Any, Any, LT]) -> None:
        assert isinstance(activelearner, ILMLBased)
        results = classifier_performance(
            activelearner.classifier, self.test_set, self.ground_truth
        )
        stats: Dict[str, float] = dict()
        labeled = activelearner.env.labeled
        for label in self.labelset:
            seen_label_docs = activelearner.env.labels.get_instances_by_label(
                label
            ).intersection(labeled)
            gen_label_docs = activelearner.env.labels.get_instances_by_label(
                label
            ).difference(labeled)
            label_performance = results[label]
            stats[f"{label}_recall"] = label_performance.recall
            stats[f"{label}_precision"] = label_performance.precision
            stats[f"{label}_f1"] = label_performance.f1
            stats[f"{label}_docs"] = len(seen_label_docs)
            stats[f"{label}_ratio"] = len(seen_label_docs) / len(labeled)
            stats[f"{label}_n_generated"] = len(gen_label_docs)
        stats["n_generated_docs"] = len(activelearner.env.all_instances) - len(
            activelearner.env.dataset
        )
        stats["n_labeled"] = len(activelearner.env.labeled)
        result_it_frame = pd.DataFrame(stats)
        self.result_frame = (
            result_it_frame
            if not self.result_frame
            else pd.concat([self.result_frame, result_it_frame])
        )

    def show(
        self, metrics: Iterable[str] = ["f1", "recall"], filename: Optional[str] = None
    ) -> None:
        # Gathering intermediate results
        df = self.result_frame
        n_labeled = df["n_labeled"]
        for metric in metrics:
            for label in self.labelset:
                metric_values = df[f"{label}_{metric}"]
                plt.plot(n_labeled, metric_values, label=f"{label} :: {metric}")
        # Plotting positive document counts
        plt.xlabel(f"number of labeled instances")
        plt.ylabel(f"metric score")
        plt.title(f"Learning curves")
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")


class MultilabelPlotter(AbstractPlotter[LT], Generic[LT]):
    def __init__(self):
        self.result_frame = pd.DataFrame()

    def update(self, activelearner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:

        stats: Dict[str, float] = dict()
        stats["dataset_size"] = len(activelearner.env.dataset)
        for label in activelearner.env.labels.labelset:
            labeled_docs = activelearner.env.labels.get_instances_by_label(
                label
            ).intersection(activelearner.env.labeled)
            stats[f"{str(label)}_count"] = len(labeled_docs)
            stats[f"{str(label)}_true_size"] = float(
                activelearner.env.truth.document_count(label)
            )
        name = name_formatter(activelearner)
        results = {**stats}
        self.result_frame = self.result_frame.append(results, ignore_index=True)

    def show(
        self,
        x_lim: Optional[float] = None,
        y_lim: Optional[float] = None,
        all_estimations: bool = False,
        filename: Optional[PathLike] = None,
    ) -> None:
        pass



from dataclasses import dataclass
import itertools
import pandas as pd
from typing import (
    FrozenSet,
    Generic,
    Optional,
    Tuple,
    List,
    Any,
)
from ..activelearning.ml_based import MLBased
from ..activelearning.ml_based import FeatureMatrix
from sklearn.linear_model import LogisticRegression
from .base import AbstractEstimator
from ..typehints.typevars import IT
from instancelib.typehints.typevars import KT, DT, RT, LT
import numpy as np
import numpy.typing as npt


@dataclass
class DecisionRow(Generic[KT]):
    key: KT
    probability: float
    labeled: bool
    pos_labeled: bool
    order: Optional[int]


class SemiEstimator(AbstractEstimator[IT, KT, DT, npt.NDArray[Any], RT, LT]):
    def semi(
        self,
        learner: MLBased[
            IT, KT, DT, npt.NDArray[Any], RT, LT, npt.NDArray[Any], npt.NDArray[Any]
        ],
        pos_label: LT,
    ) -> Tuple[float, Optional[float], Optional[float]]:
        def temporary_label_indices(y_pred_proba: npt.NDArray[Any]) -> List[int]:
            order = np.argsort(y_pred_proba)[::-1]
            print(f"{y_pred_proba[order[0]]}, {y_pred_proba[order[-1]]}")
            count = 0
            target = 1
            candidates: List[int] = []
            sample: List[int] = []
            for i, x in enumerate(y_pred_proba[order]):
                count = count + x
                candidates.append(order[i])
                if count >= target:
                    sample.append(candidates[0])
                    target = target + 1
                    candidates = []
            return sample

        env = learner.env
        classifier = learner.classifier

        def get_decision_values(pos_labeled: FrozenSet[KT], proba=True):
            pos_col_idx = classifier.get_label_column_index(pos_label)
            for mat in FeatureMatrix[KT].generator_from_provider(env.dataset, 100):
                decision_keys = mat.indices
                if proba:
                    decision_proba: List[float] = classifier.predict_proba(mat.matrix)[:, pos_col_idx].tolist()  # type: ignore
                else:
                    decision_proba: List[float] = classifier.innermodel.decision_function(mat.matrix).tolist()  # type: ignore
                decision_pos_labeled = map(
                    lambda key: key in pos_labeled, decision_keys
                )
                order = map(
                    lambda key: env.logger.get_label_order(key)
                    if key in env.labeled
                    else None,
                    decision_keys,
                )
                decision_labeled = map(lambda key: key in env.labeled, decision_keys)
                rows = itertools.starmap(
                    DecisionRow[KT],
                    zip(
                        decision_keys,
                        decision_proba,
                        decision_labeled,
                        decision_pos_labeled,
                        order,
                    ),
                )
                yield from rows

        pos_labeled = env.labels.get_instances_by_label(pos_label)
        neg_labeled = frozenset(env.labeled).difference(pos_labeled)
        pos_label_count = len(pos_labeled)
        neg_label_count = len(neg_labeled)

        decision_df = pd.DataFrame(list(get_decision_values(pos_labeled)))

        # all_idx = decision_df.index.values
        pos_idx: List[int] = decision_df.index[decision_df.pos_labeled]  # type: ignore
        # neg_idx = decision_df.index[not decision_df.pos_label]
        unl_idx = decision_df.index[decision_df.labeled == False]
        x_data = decision_df["probability"].values.reshape(-1, 1)
        y_data = decision_df["pos_labeled"].values.astype(np.int64)

        def estimate(x_data, y_data, unl_idx, neg_label_count, pos_num_last):
            regularization_strength = sum(y_data) / neg_label_count
            lreg = LogisticRegression(
                penalty="l2",
                fit_intercept=True,
                balanced=True,
                C=regularization_strength,
            )
            lreg.fit(x_data, y_data)
            pos_at = list(lreg.classes_).index(1)
            x_unl = x_data[unl_idx]
            y_pred_proba = lreg.predict_proba(x_unl)[:, pos_at]
            idx_temp_positive = temporary_label_indices(y_pred_proba)
            y_data_new = np.copy(y_data)
            for idx in unl_idx[idx_temp_positive]:
                y_data_new[idx] = 1
            pos_num = sum(y_data_new)
            if pos_num == pos_num_last:
                return pos_num
            return estimate(x_data, y_data_new, unl_idx, neg_label_count, pos_num)

        r_estimate = estimate(x_data, y_data, unl_idx, neg_label_count, 0)
        return r_estimate, 0.0, 0.0



from enum import Enum


class FECatalog:
    class DataType(Enum):
        TEXTINSTANCE = "TextInstance"
        CHATMESSAGES = "ChatMessages"
        EMAIL = "Email"
        IMAGE = "Image"

    class VectorizerType(Enum):
        STACK = "Stack"
        SKLEARN = "Sklearn"
        DOC2VEC = "Doc2Vec"
        DUALVEC = "DualVec"

    class SklearnVecType(Enum):
        TFIDF_VECTORIZER = "TfIdf"
        COUNT_VECTORIZER = "CountVectorizer"



from .base import AbstractClassifier
from .sklearn import SkLearnClassifier
from .sklearn import MultilabelSkLearnClassifier
from .factory import MachineLearningFactory
from .catalog import MachineLearningCatalog




from __future__ import annotations

import collections
import random
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)
from typing_extensions import Self

import instancelib as il
import numpy as np
import numpy.typing as npt
from instancelib.typehints import DT, KT, LT, RT, VT

from ..environment.base import AbstractEnvironment
from ..machinelearning.transductive.il_tsvm import TSVM
from ..typehints import IT
from ..utils.func import list_unzip, sort_on
from ..utils.numpy import raw_proba_chainer
from .poolbased import PoolBasedAL


class TSVMLearner(PoolBasedAL[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]):
    rank_history: Dict[int, Mapping[KT, int]]
    sampled_sets: Dict[int, Sequence[KT]]
    current_sample: Deque[KT]
    classifier: il.AbstractClassifier[
        IT, KT, DT, VT, RT, LT, npt.NDArray[Any], npt.NDArray[Any]
    ]

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        pos_label: LT,
        neg_label: LT,
        k_sample: int,
        batch_size: int,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(env, *_, identifier=identifier, **__)
        # Problem definition
        self.classifier = TSVM()
        self.pos_label = pos_label
        self.neg_label = neg_label
        self.classifier.set_target_labels([neg_label, pos_label])

        # Batch and sample sizes
        self.k_sample = k_sample
        self.batch_size = batch_size

        # Record keeping for the current sample
        self.it = 0
        self.current_sample = collections.deque()

        # Record keeping for recall analysis
        self.rank_history = dict()
        self.sampled_sets = dict()

    def __call__(
        self, environment: AbstractEnvironment[IT, KT, DT, VT, RT, LT]
    ) -> PoolBasedAL[IT, KT, DT, VT, RT, LT]:
        super().__call__(environment)
        return self

    def update_ordering(self) -> bool:
        return True

    def _provider_sample(
        self, provider: il.InstanceProvider[IT, KT, DT, VT, RT]
    ) -> il.InstanceProvider[IT, KT, DT, VT, RT]:
        k_sample = min(self.k_sample, len(provider))
        sampled_keys = random.sample(provider.key_list, k_sample)
        sampled_provider = self.env.create_bucket(sampled_keys)
        return sampled_provider

    def _temp_augment_and_train(self):
        self.classifier.fit_provider(self.env.dataset, self.env.labels)

    def _predict(
        self, provider: il.InstanceProvider[IT, KT, DT, VT, RT]
    ) -> Tuple[Sequence[KT], npt.NDArray[Any]]:
        raw_probas = self.classifier.predict_proba_provider_raw(provider)
        keys, matrix = raw_proba_chainer(raw_probas)
        return keys, matrix

    def _rank(
        self, provider: il.InstanceProvider[IT, KT, DT, VT, RT]
    ) -> Sequence[Tuple[KT, float]]:
        keys, matrix = self._predict(provider)
        pos_column = self.classifier.get_label_column_index(self.pos_label)
        prob_vector = matrix[:, pos_column]
        floats: Sequence[float] = prob_vector.tolist()
        zipped = list(zip(keys, floats))
        ranking = sort_on(1, zipped)
        return ranking

    def _sample(self, distribution: Sequence[Tuple[KT, float]]) -> Sequence[KT]:
        keys, _ = list_unzip(distribution)
        sample = keys[0 : self.batch_size]
        return sample

    @classmethod
    def _to_history(cls, ranking: Sequence[Tuple[KT, float]]) -> Mapping[KT, int]:
        keys, _ = list_unzip(ranking)
        history = {k: i for i, k in enumerate(keys, start=1)}
        return history

    def update_sample(self) -> Deque[KT]:
        if not self.current_sample:
            self._temp_augment_and_train()
            ranking = self._rank(self.env.unlabeled)
            sample = self._sample(ranking)

            # Store current sample
            self.current_sample = collections.deque(sample)

            # Record keeping
            self.rank_history[self.it] = self._to_history(ranking)
            self.sampled_sets[self.it] = tuple(sample)

            # Increment batch_size for next for train iteration
            self.batch_size += round(self.batch_size / 10)
            self.it += 1
        return self.current_sample

    def __next__(self) -> IT:
        self.update_sample()
        while self.current_sample:
            ins_key = self.current_sample.popleft()
            if ins_key not in self.env.labeled:
                return self.env.dataset[ins_key]
        if not self.env.unlabeled.empty:
            return self.__next__()
        raise StopIteration()

    @classmethod
    def builder(cls, *args: Any, **kwargs: Any) -> Callable[..., Self]:
        return super().builder(*args, **kwargs)



from multiprocessing import Process
import threading
import random
import logging
import time
from queue import Queue, Empty
from typing import Generic, TypeVar, Union, Iterator, Callable, List, Sequence
PT = TypeVar("PT")
CT = TypeVar("CT")

LOGGER = logging.getLogger(__name__)

class Sentinel(object):
    pass


def producer_func(input: Callable[..., Iterator[PT]], 
             output: Queue, 
             n_consumers) -> None:
    iterator = input()
    for item in iterator:
        output.put(item)
        print("Produced an item")
    for _ in range(n_consumers):
        output.put(Sentinel())
    return


def consumer_func(input: Queue,
             output: Queue,
             func: Callable[[PT], CT]) -> None:
    while True:
        if not input.empty():
            try:
                item = input.get()
            except Empty:
                continue
            else:
                if isinstance(item, Sentinel):
                    return
                result = func(item)
                output.put(result)
                input.task_done()
            print("Consumed an item")


class ProducerConsumer(Generic[PT, CT]):
    def __init__(self,
                 n_consumers: int = 1,
                 ):
        self.input_queue = Queue()
        self.result_queue = Queue()
        self.n_consumers = n_consumers

    def map_async(self,
                  function: Callable[[PT], Iterator[CT]],
                  iterator_builder: Callable[..., Iterator[CT]]) -> Sequence[CT]:
        producer = Process(target=producer_func, args=(iterator_builder, self.input_queue, self.n_consumers))
        producer.start()
        LOGGER.info("Producer has started")
        consumers = [
            Process(target=consumer_func, args=(self.input_queue, self.result_queue, function)) for _ in range(self.n_consumers)
        ]
        for c in consumers:
            c.start()
        LOGGER.info("Started consumers, waiting for, producer to join")
        producer.join()
        LOGGER.info("Producer has joined, waiting for consumers to join")
        for c in consumers:
            c.join()
            LOGGER.info("Consumer has joined")
        LOGGER.info("Consumers have joined")
        results: List[CT] = []
        LOGGER.info("Retrieving all info from result queue")
        while True:
            try:
                elem = self.result_queue.get()
                results.append(elem)
            except Empty:
                break
        LOGGER.info("Retrieved all info from result queue, returning results to caller")
        return results






from enum import Enum


class ALConfiguration(Enum):
    NaiveBayesEstimator = "NaiveBayesEstimator"
    SVMEstimator = "SVMEstimator"
    MixedEstimator = "MixedEstimator"
    RaschEstimator = "RaschEstimator"
    RaschLR = "RaschLR"
    RaschRF = "RaschRF"
    EntropySamplingNB = "EntropySamplingNB"
    ProbabilityEnsemble = "ProbabilityEnsemble"
    RaschNBLRRF = "RaschNBLRRF"
    RaschNBLRRFSVM = "RaschNBLRRFSVM"
    RaschNBLRRFLGBM = "RaschNBLRRFLBGM"
    RaschNBLRRFLGBMRAND = "RaschNBLRRFLGBMRAND"
    AUTOTAR = "AUTOTAR"
    AUTOSTOP = "AUTOSTOP"


class INITConfiguration(Enum):
    IDENTITY = "IDENTITY"
    RANDOM = "RANDOM"
    UNIFORM = "UNIFORM"
    SEPARATE = "SEPARATE"
    POSITVEUNIFORM = "POSITIVEUNIFORM"


class FEConfiguration(Enum):
    TFIDF5000 = "TfIDF5000"
    TFIDFTAR = "TFIDFTAR"


class EstimationConfiguration(str, Enum):
    RaschRidge = "RaschRidge"
    RaschParametric = "RaschParametric"
    RaschApproxParametric = "RaschApproxParametric"
    RaschApproxConvParametric = "RaschApproxConvParametric"
    CHAO = "Chao"
    AUTOSTOP = "AUTOSTOP"


class StopBuilderConfiguration(str, Enum):
    CHAO_CONS_OPT = "Chao_CONS_OPT"
    AUTOTAR = "AUTOTAR"
    AUTOSTOP = "AUTOSTOP"


class ExperimentCombination(str, Enum):
    CHAO4 = "CHAO4"
    AUTOTAR = "AUTOTAR"
    AUTOSTOP = "AUTOSTOP"



import collections
from abc import ABC, abstractmethod
from typing import Any, Deque, FrozenSet, Generic, Mapping, Optional, Sequence, Tuple

from instancelib import Instance

from ..activelearning.base import ActiveLearner
from ..environment import AbstractEnvironment
from ..environment.base import AbstractEnvironment
from ..typehints import DT, IT, KT, LT, RT, VT
from .base import ActiveLearner
import numpy as np


class AbstractStatistics(ABC, Generic[KT, LT]):
    @abstractmethod
    def update(self, learner: ActiveLearner[Any, KT, Any, Any, Any, LT]):
        raise NotImplementedError

    @property
    @abstractmethod
    def dataset_size(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def rounds(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def current_label_count(self, label: LT) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def current_annotated(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def label_annotated(self, label: LT, it: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def label_per_round(self, label: LT) -> Sequence[int]:
        raise NotImplementedError

    @property
    @abstractmethod
    def annotations_per_round(self) -> Sequence[int]:
        raise NotImplementedError

    def annotations_since_last(self, label: LT) -> int:
        last_pos_it = max(
            [it for it, count in enumerate(self.label_per_round(label)) if count > 0]
        )
        annotated_at = np.array(self.annotations_per_round).cumsum()[last_pos_it]
        return self.current_annotated - annotated_at


class StatsMixin(ABC, Generic[KT, LT]):
    @property
    @abstractmethod
    def stats(self) -> AbstractStatistics[KT, LT]:
        raise NotImplementedError


class StatsWrapper(StatsMixin[KT, LT], ActiveLearner[IT, KT, DT, VT, RT, LT]):
    learner: ActiveLearner[IT, KT, DT, VT, RT, LT]

    def __init__(
        self,
        learner: ActiveLearner[IT, KT, DT, VT, RT, LT],
        stats: AbstractStatistics[KT, LT],
        batch_size: int = 20,
    ):
        self.learner = learner
        self._stats = stats
        self.batch_it = batch_size
        self.batch_size = batch_size

    @property
    def stats(self) -> AbstractStatistics[KT, LT]:
        return self._stats

    def __next__(self) -> IT:
        if self.batch_it >= self.batch_size:
            self.stats.update(self.learner)
            self.learner.update_ordering()
            self.batch_it = 0
        elem = self.learner.__next__()
        self.batch_it += 1
        return elem

    @property
    def env(self) -> AbstractEnvironment[IT, KT, DT, VT, RT, LT]:
        return self.learner.env

    def update_ordering(self) -> bool:
        return self.learner.update_ordering()

    @property
    def has_ordering(self) -> bool:
        return self.learner.has_ordering

    def set_as_labeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        self.learner.set_as_labeled(instance)

    def set_as_unlabeled(self, instance: Instance[KT, DT, VT, RT]) -> None:
        self.learner.set_as_unlabeled(instance)

    @property
    def name(self) -> Tuple[str, Optional[LT]]:
        return self.learner.name


class AnnotationStatistics(AbstractStatistics[KT, LT]):
    labelwise: Deque[Mapping[LT, FrozenSet[KT]]]
    per_round: Deque[Mapping[LT, FrozenSet[KT]]]
    annotated_per_round: Deque[FrozenSet[KT]]
    annotated: Deque[FrozenSet[KT]]
    unlabeled: Deque[FrozenSet[KT]]
    dataset: Deque[FrozenSet[KT]]

    def __init__(self) -> None:
        self.labelwise = collections.deque()
        self.per_round = collections.deque()
        self.annotated = collections.deque()
        self.annotated_per_round = collections.deque()
        self.unlabeled = collections.deque()
        self.dataset = collections.deque()

    def update(self, learner: ActiveLearner[Any, KT, Any, Any, Any, LT]):
        previous_round = frozenset() if not self.annotated else self.annotated[-1]
        annotated = frozenset(learner.env.labeled)
        unlabeled = frozenset(learner.env.unlabeled)
        current_round = {
            label: (
                frozenset(
                    learner.env.get_subset_by_labels(
                        learner.env.labeled, label, labelprovider=learner.env.labels
                    )
                )
            )
            for label in learner.env.labels.labelset
        }
        current_round_new = {
            label: keys.difference(previous_round)
            for label, keys in current_round.items()
        }
        annotated_new = annotated.difference(previous_round)
        self.annotated_per_round.append(annotated_new)
        self.labelwise.append(current_round)
        self.annotated.append(annotated)
        self.unlabeled.append(unlabeled)
        self.per_round.append(current_round_new)
        self.dataset.append(frozenset(learner.env.dataset))

    @property
    def dataset_size(self) -> int:
        return 0 if not self.dataset else len(self.dataset[-1])

    @property
    def rounds(self) -> int:
        return len(self.annotated)

    def current_label_count(self, label: LT) -> int:
        return 0 if not self.labelwise else len(self.labelwise[-1][label])

    @property
    def current_annotated(self) -> int:
        return 0 if not self.annotated else len(self.annotated[-1])

    def label_annotated(self, label: LT, it: int) -> int:
        return 0 if not self.per_round else len(self.per_round[it][label])

    def label_per_round(self, label: LT) -> Sequence[int]:
        return [len(dc[label]) for dc in self.per_round]

    @property
    def annotations_per_round(self) -> Sequence[int]:
        return [len(an) for an in self.annotated_per_round]



from abc import ABC, abstractmethod
from typing import (Any, FrozenSet, Generic, Iterable, Sequence,
                    TypeVar, Union)

from instancelib.instances import Instance
from instancelib.labels.base import LabelProvider
from instancelib.typehints import KT, LT

ST = TypeVar("ST")
class BaseLogger(ABC, Generic[KT, LT, ST]):
    @abstractmethod
    def log_ordering(self, ordering: Sequence[KT], metrics: Sequence[float], labeled: Iterable[KT], labels: LabelProvider[KT, LT]) -> None:
        raise NotImplementedError

    @abstractmethod
    def log_sample(self, x: Union[KT, Instance[KT, Any, Any, Any]], sample_method: ST) -> None:
        raise NotImplementedError

    @abstractmethod
    def log_label(self, x: Union[KT, Instance[KT, Any, Any, Any]], sample_method: ST,  *labels: LT) -> None:
        raise NotImplementedError

    @abstractmethod
    def get_sampled_info(self, x: Union[KT, Instance[KT, Any, Any, Any]]) -> FrozenSet[ST]:
        raise NotImplementedError

    @abstractmethod
    def get_instances_by_method(self, sample_method: ST) -> FrozenSet[KT]:
        raise NotImplementedError

    @abstractmethod
    def get_label_order(self, x: Union[KT, Instance[KT, Any, Any, Any]]) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def labelset(self) -> FrozenSet[LT]: 
        raise NotImplementedError



from __future__ import annotations

import collections
import itertools
import logging
import os
from abc import ABC, abstractmethod
from typing import (
    Any,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)
import warnings

import numpy as np
import numpy.typing as npt

import pandas as pd  # type: ignore

from ..activelearning.base import ActiveLearner
from ..activelearning.estimator import Estimator
from ..utils.func import (
    all_subsets,
    intersection,
    list_unzip3,
    not_in_supersets,
    powerset,
    union,
)
from .base import AbstractEstimator, Estimate

try:
    import rpy2.robjects as ro  # type: ignore
    from rpy2.robjects import pandas2ri  # type: ignore
    from rpy2.robjects.conversion import localconverter  # type: ignore
    from rpy2.robjects.packages import importr  # type: ignore
except ImportError:
    R_AVAILABLE = False
else:
    R_AVAILABLE = True

from ..typehints import IT, KT, DT, RT, LT, VT

LOGGER = logging.getLogger(__name__)

from ..analysis.statistics import EstimationModelStatistics


def _check_R():
    """Checks if Python <-> R interop is available

    Raises
    ------
    ImportError
        If the interop is not available
    """
    if not R_AVAILABLE:
        raise ImportError("Install rpy2 interop")


class AbundanceEstimator(
    AbstractEstimator[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    r_loaded: bool
    name = "Mh Chao Estimator"

    def __init__(self):
        self.matrix_history: Deque[pd.DataFrame] = collections.deque()
        self.contingency_history: Deque[Dict[FrozenSet[int], int]] = collections.deque()
        self.r_loaded = False

    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "mhmodel.R")
        R["source"](r_script_file)

    def get_label_matrix(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> pd.DataFrame:
        rows = {
            ins_key: {
                l_key: ins_key in learner.env.labeled
                for l_key, learner in enumerate(estimator.learners)
            }
            for ins_key in estimator.env.labels.get_instances_by_label(label)
        }
        dataframe = pd.DataFrame.from_dict(rows, orient="index")  # type: ignore
        self.matrix_history.append(dataframe)
        return dataframe

    def get_contingency_list(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> Dict[FrozenSet[int], int]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(label).intersection(
                learner.env.labeled
            )
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: len(intersection(*[learner_sets[key] for key in combination]))
            for combination in key_combinations
            if len(combination) >= 1
        }
        self.contingency_history.append(result)
        return result

    def get_matrix(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> npt.NDArray[Any]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(label).intersection(
                learner.env.labeled
            )
            for learner_key, learner in enumerate(estimator.learners)
        }
        n_learners = len(learner_sets)
        matrix = np.zeros(shape=(n_learners, n_learners))
        for i, key_a in enumerate(learner_sets):
            instances_a = learner_sets[key_a]
            for j, key_b in enumerate(learner_sets):
                if i != j:
                    instances_b = learner_sets[key_b]
                    intersection = instances_a.intersection(instances_b)
                    matrix[i, j] = len(intersection)
        return matrix

    def calculate_abundance_R(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> pd.DataFrame:
        if not self.r_loaded:
            self._start_r()
            self.r_loaded = True
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            df = self.get_label_matrix(estimator, label)
            with localconverter(ro.default_converter + pandas2ri.converter):
                df_r = ro.conversion.py2rpy(df)
                abundance_r = ro.globalenv["get_abundance"]
                r_df = abundance_r(df_r)
                res_df: pd.DataFrame = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_abundance(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> Estimate:
        res_df = self.calculate_abundance_R(estimator, label)

        def try_float(val: Any) -> float:
            try:
                parsed = float(val)
            except ValueError:
                parsed = float("nan")
            return parsed

        point = try_float(res_df["abundance"][0])
        lower = try_float(res_df["infCL"][0])
        upper = try_float(res_df["supCL"][0])
        return Estimate(point, lower, upper)

    def __call__(
        self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT
    ) -> Estimate:
        empty = np.array([])
        stats = EstimationModelStatistics(empty, empty, 0.0, empty)
        if not isinstance(learner, Estimator):
            return Estimate(0.0, 0.0, 0.0)
        estimate = self.calculate_abundance(learner, label)
        return estimate

    def all_estimations(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> Sequence[Tuple[str, float, float]]:
        res_df = self.calculate_abundance_R(estimator, label)
        ok_fit = res_df[res_df.infoFit == 0]
        if len(ok_fit) == 0:
            ok_fit = res_df
        results = ok_fit.values
        names = list(ok_fit.index)
        estimations = list(results[:, 0])
        errors = list(results[:, 1])
        tuples = list(zip(names, estimations, errors))
        return tuples

    def get_contingency_sets(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> Dict[FrozenSet[int], FrozenSet[KT]]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(label).intersection(
                learner.env.labeled
            )
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: intersection(*[learner_sets[key] for key in combination])
            for combination in key_combinations
            if len(combination) >= 1
        }
        filtered_result = not_in_supersets(result)
        return filtered_result

    def get_occasion_history(
        self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT
    ) -> pd.DataFrame:
        contingency_sets = self.get_contingency_sets(estimator, label)
        learner_keys = union(*contingency_sets.keys())
        rows = {
            i: {
                **{
                    f"learner_{learner_key}": int(learner_key in combination)
                    for learner_key in learner_keys
                },
                **{"count": len(instances)},
            }
            for (i, (combination, instances)) in enumerate(contingency_sets.items())
        }
        df = pd.DataFrame.from_dict(rows, orient="index")  # type: ignore
        return df



# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import pathlib
import setuptools  # type: ignore

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setuptools.setup(  # type: ignore
    name="python-allib",
    version="0.2.0.0",  # NOSONAR
    description="A typed Active Learning Library",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Michiel Bron",
    author_email="m.p.bron@uu.nl",
    license="GNU LGPL v3",
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    package_data={"allib": ["py.typed", "estimation/mhmodel.R"]},
    packages=setuptools.find_packages(),  # type: ignore
    python_requires=">=3.8",
    install_requires=[
        "numba",
        "numpy",
        "pandas",
        "h5py",
        "tables",
        "scikit-learn",
        "scipy",
        "openpyxl",
        "xlrd",
        "instancelib",
        "imblearn",
        "lightgbm",
        "gensim",
        "more-itertools",
        "matplotlib",
        "typing_extensions>=4.4.0",
    ],
    extras_require={"doc2vec": ["gensim"]},
)



import typing as ty
from abc import ABC, abstractmethod
from collections import OrderedDict
from os import PathLike
from typing import (Any, Dict, FrozenSet, Generic, Iterable, List, Mapping,
                    Optional, Sequence, Tuple, TypeVar, Union)

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from ..analysis.statistics import EstimationModelStatistics
from ..estimation.base import AbstractEstimator, Estimate
from ..typehints import LT
from .experiments import ExperimentIterator
from .plotter import ExperimentPlotter
from .statistics import TarDatasetStats, TemporalRecallStats

_T = TypeVar("_T")
_V = TypeVar("_V")
_W = TypeVar("_W")


def smooth_similar(
    xs: Union[npt.NDArray[Any], Sequence[_T]], ys: Union[npt.NDArray[Any], Sequence[_V]]
) -> Tuple[Sequence[_T], Sequence[_V]]:
    assert len(xs) == len(ys), f"len(xs) != len(ys); {len(xs)} !=  {len(ys)}"
    x_smoothed: List[_T] = list()
    y_smoothed: List[_V] = list()
    previous_y: Optional[_V] = None
    for x, y in zip(xs, ys):
        if previous_y != y:
            x_smoothed.append(x)
            y_smoothed.append(y)
            previous_y = y
    return x_smoothed, y_smoothed


def smooth_similar3(
    xs: Union[npt.NDArray[Any], Sequence[_T]],
    ys: Union[npt.NDArray[Any], Sequence[_V]],
    zs: Union[npt.NDArray[Any], Sequence[_W]],
) -> Tuple[Sequence[_T], Sequence[_V], Sequence[_W]]:
    assert (
        len(xs) == len(ys) == len(zs)
    ), f"len(xs) != len(ys) != len(zs); {len(xs)} !=  {len(ys)} != {len(zs)}"
    x_smoothed: List[_T] = list()
    y_smoothed: List[_V] = list()
    z_smoothed: List[_W] = list()
    previous_y: Optional[_V] = None
    previous_z: Optional[_W] = None
    for x, y, z in zip(xs, ys, zs):
        if previous_y != y or previous_z != z:
            x_smoothed.append(x)
            y_smoothed.append(y)
            z_smoothed.append(z)
            previous_y = y
            previous_z = z
    return x_smoothed, y_smoothed, z_smoothed


class TarExperimentPlotter(ExperimentPlotter[LT], Generic[LT]):
    pos_label: LT
    neg_label: LT

    dataset_name: str

    dataset_stats: ty.OrderedDict[int, TarDatasetStats]
    recall_stats: ty.OrderedDict[int, TemporalRecallStats]
    estimates: ty.OrderedDict[int, Mapping[str, Estimate]]
    stop_results: ty.OrderedDict[int, Mapping[str, bool]]

    def __init__(self, pos_label: LT, neg_label: LT, dataset_name: str = "") -> None:
        self.pos_label = pos_label
        self.neg_label = neg_label

        self.dataset_name = dataset_name

        self.dataset_stats = OrderedDict()
        self.recall_stats = OrderedDict()
        self.estimates = OrderedDict()
        self.stop_results = OrderedDict()
        self.it = 0
        self.it_axis: List[int] = list()

    def update(
        self,
        exp_iterator: ExperimentIterator[Any, Any, Any, Any, Any, LT],
        stop_result: Mapping[str, bool],
    ) -> None:
        learner = exp_iterator.learner
        self.it = exp_iterator.it
        self.it_axis.append(self.it)
        self.recall_stats[self.it] = TemporalRecallStats.from_learner(
            learner, self.pos_label, self.neg_label
        )
        self.dataset_stats[self.it] = TarDatasetStats.from_learner(
            learner, self.pos_label, self.neg_label
        )
        self.estimates[self.it] = {
            name: estimate for name, estimate in exp_iterator.recall_estimate.items()
        }
        self.stop_results[self.it] = stop_result

    @property
    def estimator_names(self) -> FrozenSet[str]:
        if self.estimates:
            return frozenset(self.estimates[self.it].keys())
        return frozenset()

    @property
    def criterion_names(self) -> FrozenSet[str]:
        if self.stop_results:
            return frozenset(self.stop_results[self.it].keys())
        return frozenset()

    def _effort_axis(self) -> npt.NDArray[Any]:
        effort_axis = np.array([self.recall_stats[it].effort for it in self.it_axis])
        return effort_axis

    def exp_random_recall(self, it: int) -> float:
        effort = self.recall_stats[it].effort
        dataset_size = self.dataset_stats[it].size
        true_pos = self.dataset_stats[it].pos_count
        expected = effort / dataset_size * true_pos
        return expected

    def print_last_stats(self) -> None:
        estimate = self.estimates[self.it]
        recall = self.recall_stats[self.it]
        print(estimate)
        print(recall.pos_docs_found)

    def plot_recall_statistic(
        self, stats: Mapping[int, TemporalRecallStats], key: str, label: str
    ) -> None:
        effort_axis = self._effort_axis()
        curve = np.array([stats[it].__dict__[key] for it in self.it_axis])
        plt.plot(effort_axis, curve, label=label)

    def _plot_estimator(self, key: str, color="gray", alpha=0.2) -> None:
        effort_axis = self._effort_axis()
        points = np.array([self.estimates[it][key].point for it in self.it_axis])
        lows = np.array([self.estimates[it][key].lower_bound for it in self.it_axis])
        uppers = np.array([self.estimates[it][key].upper_bound for it in self.it_axis])
        xs, ys = smooth_similar(effort_axis, points)
        xrs, ls, us = smooth_similar3(effort_axis, lows, uppers)
        plt.plot(xs, ys, linestyle="-.", label=f"Estimate by {key}", color=color)
        plt.fill_between(xrs, ls, us, color=color, alpha=alpha)  # type: ignore

    def _plot_stop_criteria(self) -> None:
        results: Sequence[Tuple[int, float, float, str, str]] = list()
        for i, crit_name in enumerate(self.criterion_names):
            color = f"C{i}"
            for it in self.it_axis:
                frame = self.stop_results[it]
                if frame[crit_name]:
                    wss = self.recall_stats[it].wss
                    recall = self.recall_stats[it].recall
                    results.append((it, recall, wss, crit_name, color))
                    break
        results_sorted = sorted(results)
        for it, recall, wss, crit_name, color in results_sorted:
            exp_found = self.exp_random_recall(it)
            act_found = self.recall_stats[it].pos_docs_found
            plt.vlines(
                x=self.recall_stats[it].effort,
                ymin=exp_found,
                ymax=act_found,
                linestyles="dashed",
                color=color,
                label=f"{crit_name} WSS: {(wss*100):.1f} %, "
                f"Recall: {(recall*100):.1f} %",
            )

    def _graph_setup(self, simulation=True) -> None:
        true_pos = self.dataset_stats[self.it].pos_count
        dataset_size = self.dataset_stats[self.it].size

        plt.xlabel(f"number of read documents")
        plt.ylabel("number of retrieved relevant documents")
        if simulation:
            plt.title(
                f"Run on a dataset with {int(true_pos)} inclusions out of {int(dataset_size)}"
            )
        else:
            plt.title(f"Run on a dataset of {int(dataset_size)}")

    def _plot_static_data(self, recall_target: float) -> None:
        # Static data
        true_pos = self.dataset_stats[self.it].pos_count
        dataset_size = self.dataset_stats[self.it].size

        pos_target = int(np.ceil(recall_target * true_pos))
        effort_axis = self._effort_axis()

        plt.axhline(y=true_pos, linestyle=":", label=f"100 % recall ({true_pos})")
        plt.axhline(
            y=pos_target,
            linestyle=":",
            label=f"{int(recall_target * 100)} % recall ({pos_target})",
        )
        plt.plot(
            effort_axis,
            (effort_axis / dataset_size) * true_pos,
            ":",
            label=f"Exp. found at random",
        )

    def _plot_recall_stats(self, included: Optional[Sequence[str]] = list()) -> None:
        # Gather and reorganize recall data
        recall_stats = TemporalRecallStats.transpose_dict(self.recall_stats)
        # Plot pos docs docs found
        for name, stats in recall_stats.items():
            if included is None or name in included:
                self.plot_recall_statistic(
                    stats, "pos_docs_found", f"# found by {name}"
                )

    def _plot_estimators(
        self, included_estimators: Optional[Sequence[str]] = None
    ) -> None:
        if included_estimators is None:
            included_estimators = list(self.estimator_names)
        # Plotting estimations
        for i, estimator in enumerate(included_estimators):
            self._plot_estimator(estimator, color=f"C{i}")

    def _set_axes(
        self, x_lim: Optional[float] = None, y_lim: Optional[float] = None
    ) -> None:
        # Setting axis limitations
        true_pos = self.dataset_stats[self.it].pos_count
        if x_lim is not None:
            plt.xlim(0, x_lim)
        if y_lim is not None:
            plt.ylim(0, y_lim)
        else:
            plt.ylim(0, 1.4 * true_pos)

    def show(
        self,
        x_lim: Optional[float] = None,
        y_lim: Optional[float] = None,
        recall_target: float = 0.95,
        included_estimators: Optional[Sequence[str]] = None,
        included_models: Optional[Sequence[str]] = None,
        filename: "Optional[PathLike[str]]" = None,
    ) -> None:
        self._graph_setup()
        self._plot_static_data(recall_target)
        self._plot_recall_stats(included_models)
        self._plot_estimators(included_estimators)
        self._plot_stop_criteria()
        self._set_axes(x_lim, y_lim)
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left")

        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
            plt.close()

    def user_show(
        self,
        x_lim: Optional[float] = None,
        y_lim: Optional[float] = None,
        included_estimators: Optional[Sequence[str]] = None,
        included_models: Optional[Sequence[str]] = None,
        filename: "Optional[PathLike[str]]" = None,
    ):
        self._graph_setup(simulation=False)
        self._plot_recall_stats(included_models)
        self._plot_estimators(included_estimators)
        self._plot_stop_criteria()
        self._set_axes(x_lim, y_lim)
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")
            plt.close()

    def wss_at_target(self, target: float) -> float:
        for it in self.it_axis:
            frame = self.recall_stats[it]
            if frame.recall >= target:
                return frame.wss
        return float("nan")

    def recall_at_stop(self, stop_criterion: str) -> float:
        stop_it = self._it_at_stop(stop_criterion)
        if stop_it is not None:
            return self.recall_stats[stop_it].recall
        return float("nan")

    def _it_at_stop(self, stop_criterion: str) -> Optional[int]:
        for it in self.it_axis:
            frame = self.stop_results[it]
            if frame[stop_criterion]:
                return it
        return None

    def wss_at_stop(self, stop_criterion: str) -> float:
        stop_it = self._it_at_stop(stop_criterion)
        if stop_it is not None:
            return self.recall_stats[stop_it].wss
        return float("nan")

    def relative_error(self, stop_criterion: str, recall_target: float) -> float:
        recall_at_stop = self.recall_at_stop(stop_criterion)
        relative_error = abs(recall_at_stop - recall_target) / recall_target
        return relative_error

    def loss_er_at_stop(self, stop_criterion: str) -> float:
        stop_it = self._it_at_stop(stop_criterion)
        if stop_it is not None:
            return self.recall_stats[stop_it].loss_er
        return float("nan")

    def effort_at_stop(self, stop_criterion: str) -> int:
        stop_it = self._it_at_stop(stop_criterion)
        if stop_it is not None:
            return self.recall_stats[stop_it].effort
        return self.recall_stats[self.it_axis[-1]].effort

    def proportional_effort_at_stop(self, stop_criterion: str) -> float:
        stop_it = self._it_at_stop(stop_criterion)
        if stop_it is not None:
            return self.recall_stats[stop_it].proportional_effort
        return 1.0


def filter_model_infos(
    mapping: Mapping[str, AbstractEstimator]
) -> Mapping[str, EstimationModelStatistics]:
    results = {
        key: est.model_info[-1]  # type: ignore
        for key, est in mapping.items()  # type: ignore
        if hasattr(est, "model_info")
    }  # type: ignore
    return results


class ModelStatsTar(TarExperimentPlotter[LT]):
    model_stats: ty.OrderedDict[int, Mapping[str, EstimationModelStatistics]]

    def __init__(self, pos_label: LT, neg_label: LT, dataset_name: str = "") -> None:
        super().__init__(pos_label, neg_label, dataset_name)
        self.model_stats = OrderedDict()

    def update(
        self,
        exp_iterator: ExperimentIterator[Any, Any, Any, Any, Any, LT],
        stop_result: Mapping[str, bool],
    ) -> None:
        super().update(exp_iterator, stop_result)
        self.model_stats[self.it] = filter_model_infos(exp_iterator.estimators)

    def _plot_estimators(
        self, included_estimators: Optional[Sequence[str]] = None
    ) -> None:
        super()._plot_estimators(included_estimators)
        if included_estimators is None:
            included_estimators = list(self.estimator_names)
        assert included_estimators is not None
        for estimator in included_estimators:
            if estimator in self.model_stats[self.it]:
                deviances = [
                    self.model_stats[it][estimator].deviance for it in self.it_axis
                ]
                plt.plot(
                    self._effort_axis(),
                    deviances,
                    label=f"Deviance {estimator}",
                    linestyle="--",
                )



import collections
import functools
from typing import Any, Deque, FrozenSet, Generic, Mapping, Tuple

import pandas as pd
import numpy as np
import numpy.typing as npt
from instancelib.typehints import DT, KT, LT, RT, VT

from string import ascii_uppercase
from ..activelearning import ActiveLearner
from ..activelearning.estimator import Estimator
from ..utils.func import all_subsets, intersection, not_in_supersets, powerset, union
from .base import AbstractEstimator, Estimate
from .rasch import EMRaschCombined
from .rasch_multiple import pos_rasch_numpy
from ..analysis.statistics import EstimationModelStatistics
from ..typehints import IT


def log_linear_estimate(df: pd.DataFrame, dataset_size: int, multinomial_size: int) -> Tuple[Estimate, EstimationModelStatistics]:
    df_formatted = df.sort_values([col for col in ascii_uppercase if col in df.columns])
    df_formatted.insert(0, "intercept", 1)
    design_mat = (
        df_formatted.loc[:, df_formatted.columns != 'count'] # type: ignore
            .values) # type: ignore
    design_mat = design_mat.astype("float64")
    obs_counts: npt.NDArray[Any] = df_formatted["count"].values # type: ignore
    total_found: int = int(np.sum(obs_counts))
    
    point, low, up, beta, mfit, deviance, estimates = pos_rasch_numpy(
        design_mat, obs_counts, total_found, max_it=multinomial_size)
    estimate = Estimate(point, low, up)
    stats = EstimationModelStatistics(beta, mfit, deviance, estimates)
    return estimate, stats

def log_linear_em_estimate(df: pd.DataFrame, dataset_size: int, multinomial_size: int) -> Tuple[Estimate, EstimationModelStatistics]:
    df_formatted = df.sort_values([col for col in ascii_uppercase if col in df.columns])
    df_formatted.insert(0, "intercept", 1)
    design_mat = (
        df_formatted.loc[:, df_formatted.columns != 'count'] # type: ignore
            .values) # type: ignore
    design_mat = design_mat.astype("float64")
    obs_counts: npt.NDArray[Any] = df_formatted["count"].values # type: ignore
    total_found: int = int(np.sum(obs_counts))
    
    point, low, up, beta, mfit, deviance, estimates = pos_rasch_numpy(
        design_mat, obs_counts, total_found, max_it=multinomial_size)
    estimate = Estimate(point, low, up)
    stats = EstimationModelStatistics(beta, mfit, deviance, estimates)
    return estimate, stats

def learner_key(l: int) -> str:
    return ascii_uppercase[l] 

def learnercombi_string(lset: FrozenSet[int]) -> str:
    strlset = "".join([learner_key(l) for l in sorted(lset)])
    return strlset

def get_contingency_sets_negative(estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                                  label: LT) -> Mapping[FrozenSet[int], FrozenSet[KT]]:
    learner_sets = {
            learner_key: (
                frozenset(learner.env.labeled).
                    difference(
                        learner.env.labels.get_instances_by_label(label))
            )
        for learner_key, learner in enumerate(estimator.learners)
    }
    key_combinations = powerset(range(len(estimator.learners)))
    result = {
        combination: intersection(*[learner_sets[key] for key in combination])
        for combination in key_combinations
        if len(combination) >= 1
    }
    filtered_result = not_in_supersets(result)
    return filtered_result

def get_contingency_sets(estimator: Estimator[Any, KT, Any, Any, Any, LT], 
                         label: LT) -> Mapping[FrozenSet[int], FrozenSet[KT]]:
    learner_sets = {
        learner_key: learner.env.labels.get_instances_by_label(
            label).intersection(learner.env.labeled)
        for learner_key, learner in enumerate(estimator.learners)
    }
    key_combinations = powerset(range(len(estimator.learners)))
    result = {
        combination: intersection(*[learner_sets[key] for key in combination])
        for combination in key_combinations
        if len(combination) >= 1
    }
    filtered_result = not_in_supersets(result)
    return filtered_result

    

class LogLinear(AbstractEstimator[IT, KT, DT, VT, RT, LT],
        Generic[IT, KT, DT, VT, RT, LT]):

    estimates: Deque[Estimate]
    model_info: Deque[EstimationModelStatistics]
    dfs: Deque[pd.DataFrame]

    def __init__(self, multinomial_size: int = 2000):
        super().__init__()
        self.estimates = collections.deque()
        self.dfs = collections.deque()
        self.model_info = collections.deque()
        self.multinomial_size = multinomial_size

    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Estimate:
        assert isinstance(learner, Estimator)
        return self.calculate_estimate(learner, label)

    def calculate_estimate(self,
                           estimator: Estimator[Any, KT, DT, VT, RT, LT],
                           label: LT) -> Estimate:
        dataset_size = len(estimator.env.dataset)
        df = self.get_design_matrix(estimator, label)
        if not self.dfs or not self.dfs[-1].equals(df):
            self.dfs.append(df)
            est, stats = log_linear_estimate(df, dataset_size, self.multinomial_size)
            self.estimates.append(est)
            self.model_info.append(stats)
        return self.estimates[-1]
    
    
    @classmethod
    def design_matrix_row(cls,
                          combination: Tuple[FrozenSet[int], FrozenSet[Any]],
                          all_learners: FrozenSet[int]) -> Mapping[str, int]:
        learner_set, instances = combination
        learner_cols = {
            learner_key(l): int(l in learner_set) 
            for l in all_learners
        }
        count_col = {"count": len(instances)}
        
        all_possible_interactions = all_subsets(all_learners, 2 , len(all_learners) -1)
        all_present_interactions = all_subsets(learner_set, 2 , len(all_learners) - 1)
        interaction_cols = {
            learnercombi_string(lset): int(lset in all_present_interactions)
            for lset in all_possible_interactions
        }
        final_row = {
            **learner_cols,
            **interaction_cols,
            **count_col
        }
        
        return final_row
    
    @classmethod
    def get_design_matrix(cls, 
                          estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                          label: LT) -> pd.DataFrame:
        contingency_sets_pos = get_contingency_sets(estimator, label)
        learner_keys = union(*contingency_sets_pos.keys())
        
        pos_func = functools.partial(
            cls.design_matrix_row, all_learners=learner_keys)
        rows = dict(enumerate(map(pos_func, contingency_sets_pos.items())))
        df = pd.DataFrame.from_dict(rows, orient="index")
        return df

    
    


class PosNegLogLinear(LogLinear):

    @classmethod
    def pos_neg_row(cls,
                    combination: Tuple[FrozenSet[int], FrozenSet[Any]],
                    all_learners: FrozenSet[int], positive: bool) -> Mapping[str, int]:
        learner_set, _ = combination
        standard_cols = cls.design_matrix_row(combination, all_learners)
        positive_col = {"positive": int(positive)}
        all_possible_interactions = all_subsets(all_learners, 2 , len(all_learners) -1)
        all_present_interactions = all_subsets(learner_set, 2 , len(all_learners) - 1)
        pos_learner_cols = {
            f"{learner_key(l)}_pos": (int(l in learner_set)  if positive else 0)
            for l in all_learners
        }
        interaction_pos_cols = {
            f"{learnercombi_string(lset)}": (int(lset in all_present_interactions) if positive else 0)
            for lset in all_possible_interactions
        }
        final_row = {
            **standard_cols,
            **positive_col,
            **pos_learner_cols,
            **interaction_pos_cols
        }
        return final_row

    @classmethod
    def get_design_matrix(cls, 
                          estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                          label: LT) -> pd.DataFrame:
        contingency_sets_pos = get_contingency_sets(estimator, label)
        contingency_sets_neg = get_contingency_sets_negative(estimator, label)
        learner_keys = union(*contingency_sets_pos.keys())
        
        pos_func = functools.partial(
            cls.pos_neg_row, all_learners=learner_keys, positive = True)
        
        neg_func = functools.partial(
            cls.pos_neg_row, all_learners=learner_keys, positive = False)
                
        rows_pos = list(map(pos_func, contingency_sets_pos.items()))
        rows_neg = list(map(neg_func, contingency_sets_neg.items()))
        rows = dict(enumerate(rows_pos + rows_neg))
        
        df = pd.DataFrame.from_dict(rows, orient="index")
        return df

    



# Copyright 2019 The ASReview Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Optional

from math import ceil

import numpy as np # type: ignore

from .base import BaseBalancer
from ..utils import get_random_generator

class UndersampleBalancer(BaseBalancer):
    """Balancing class that undersamples the data with a given ratio.
    """
    name = "Undersample"

    def __init__(self, ratio=1.0, rng: Optional[np.random.Generator] = None):
        """Initialize the undersampling balance strategy.
        Arguments
        ---------
        ratio: double
            Undersampling ratio of the zero's. If for example we set a ratio of
            0.25, we would sample only a quarter of the zeros and all the ones.
        """
        self.ratio = ratio
        self._rng = get_random_generator(rng)

    def resample(self, x_data, y_data):
        # Find all documents where that has a label:
        # This assumes that negative examples are 
        # encoded as zeroes. Works for multiclass,
        # multilabel. Works for binary classification
        # if the positive class is encoded as 1.
        if y_data.ndim > 1:
            positive_indices = np.where(np.any(y_data == 1, axis=1))[0]
            negative_indices = np.where(np.all(y_data == 0, axis=1))[0]
        else:
            positive_indices = np.where(y_data == 1)[0]
            negative_indices = np.where(y_data == 0)[0]

        n_positive = len(positive_indices)
        n_negative = len(negative_indices)

        # Calculate the ratio between positive and 
        # negative
        # If we don't have an excess of negative examples,
        # give back all training_samples.
        if n_positive / n_negative >= self.ratio:
            sample_ind = np.append(
                positive_indices, negative_indices)
        else:
            n_negative_sample = ceil(n_positive / self.ratio)
            negative_sample = self._rng.choice(
                np.arange(n_negative), n_negative_sample, replace=False)
            sample_ind = np.append(positive_indices, 
            negative_indices[negative_sample])
        shuf_ind = self._rng.permutation(sample_ind)
        return x_data[shuf_ind], y_data[shuf_ind]

    


from enum import Enum
from ..activelearning import ALCatalog
from ..feature_extraction import FECatalog
from ..balancing import BalancerCatalog 
from ..environment import EnvironmentCatalog
from ..machinelearning import MachineLearningCatalog

class ModuleCatalog:
    class AL(ALCatalog):
        pass
    class FE(FECatalog):
        pass
    class BL(BalancerCatalog):
        pass
    class ENV(EnvironmentCatalog):
        pass
    class ML(MachineLearningCatalog):
        pass


class ActiveLearningTasks(Enum):
    TAR = "TAR"
    CLASSIFICATION = "CLASS"


from __future__ import annotations

import collections
import functools
import itertools
import logging
from abc import ABC, abstractmethod
from multiprocessing import Pool
from queue import Queue
from typing import (Any, Callable, Dict, FrozenSet, Generic, Iterable, Iterator, List,
                    Optional, Sequence, Set, Tuple, TypeVar)

import numpy as np  # type: ignore
import numpy.typing as npt
from sklearn.exceptions import NotFittedError  # type: ignore

from ..environment import AbstractEnvironment
from ..exceptions import NoOrderingException, NotInitializedException
from ..exceptions.base import NoLabeledDataException, NoVectorsException
from instancelib.instances.base import Instance, InstanceProvider
from ..machinelearning import AbstractClassifier
from ..utils import divide_sequence, mapsnd
from ..utils.func import filter_snd_none, fst, list_unzip, multisort, sort_on
from ..utils.random import get_random_generator
from .base import ActiveLearner, ProbabilityPrediction
from .ensembles import AbstractEnsemble
from .selectioncriterion import AbstractSelectionCriterion
from .ml_based import (FeatureMatrix, MLBased,
                       ProbabilityBased)
from .fixed import FixedOrdering
from .random import RandomSampling

from ..typehints import KT, DT, VT, RT, LT, LVT, PVT, IT

FT = TypeVar("FT")


LOGGER = logging.getLogger(__name__)

def get_probabilities(probabilities: Optional[Sequence[float]], learners: Sequence[Any]) -> Sequence[float]:
    if len(learners) == 0:
        return []
    if probabilities is not None and probabilities:
        return probabilities
    equal_prob = 1.0 / len(learners)
    probs = [equal_prob] * len(learners)
    return probs
    



class ProbabilityBasedEnsemble(AbstractEnsemble[IT, KT, DT, npt.NDArray[Any], RT, LT], 
                       ProbabilityBased[IT, KT, DT, RT, LT], 
                       Generic[IT, KT, DT, RT, LT]):
    _name = "ProbabilityBasedEnsemble"

    def __init__(self, 
                 classifier: AbstractClassifier[KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]],
                 strategies: Sequence[AbstractSelectionCriterion],
                 probabilities: Optional[Sequence[float]] = None, rng: Any = None,
                 batch_size = 200,
                 identifier: Optional[str] = None, 
                 fallback = RandomSampling[IT, KT, DT, npt.NDArray[Any], RT, LT].builder(), *_, **__) -> None:
        self.classifier = classifier
        self.strategies = strategies
        self.learners = [FixedOrdering(identifier=strategy.name) for strategy in self.strategies]
        self.probabilities = get_probabilities(probabilities, strategies)
        self._rng: Any = get_random_generator(rng)
        self.fallback = fallback
        self.sampled: Set[KT] = set()
        self._has_ordering: bool = False
        self.batch_size = batch_size
        self.identifier = identifier
    
    def __call__(self, environment: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT]):
        """Initialize the learner with an environment

        Args:
            environment (AbstractEnvironment[KT, DT, VT, RT, LT]): the chosen environment

        Returns:
            StrategyEnsemble: The initialized environment
        """        
        super().__call__(environment)
        self.classifier = self.classifier(self.env)
        for learner in self.learners:
            learner(self.env)
        return self

    def process_predictions(self, 
                            predictions: Tuple[Sequence[KT], npt.NDArray[Any]]
                            ) -> Tuple[Sequence[KT], npt.NDArray[Any]]:
        keys, matrix = predictions
        results = np.transpose(np.array([sel_criterion(matrix) for sel_criterion in self.strategies]))
        return keys, results

    def calculate_ordering(self) -> Sequence[Tuple[Sequence[KT], Sequence[float]]]:
        """Calculate the ordering for the unlabeled set of instances according
        to their predicted label probabilities by the learner's classifier.
        The provided selection criterion calculates a metric on the probabilities.

        Returns
        -------
        Tuple[Sequence[KT], Sequence[float]]
            A tuple of two lists of equal length and ordered descendingly
            according to the second list.

            - A list of instance keys with type KT
            - A list of scores, matching with the instance keys
        """        
               
        def to_tuples(pair: Tuple[Sequence[KT], npt.NDArray[Any]]
                     ) -> Sequence[Tuple[KT, Sequence[float]]]:
            keys, matrix = pair
            results: Sequence[Sequence[float]] = [[]]
            if len(matrix.shape) < 2:
                results = matrix.reshape((len(keys), 1)).tolist()
            if len(matrix.shape) == 2:
                results = matrix.tolist()
            else:
                raise ValueError(
                    "The result matrix is incorrectly shaped. "
                    f"It should have shape ({len(keys)}, {len(self.strategies)})"
                    f"while the matrix has shape {matrix.shape}")            
            if len(keys) != len(results):
                raise ValueError(
                    "The results are incorrectly shaped." 
                    "The length of the list of keys and the list of scores should be equal.")
            zipped = list(zip(keys, results))
            return zipped

        LOGGER.info("[%s] Start calculating the prediction matrix for all unlabeled documents", self.name)
        # Get a generator with that generates feature matrices from data
        matrices = FeatureMatrix[KT].generator_from_provider(
            self.env.unlabeled, self.batch_size)
        # Get the predictions for each matrix
        predictions = map(self._get_predictions, matrices)
        # Apply selections strategies on the predictions
        metric_results = map(self.process_predictions, predictions)
    
        # Convert each Tuple[Sequence[KT], npt.NDArray[Any]] with the metric results to a list 
        metric_tuples = map(to_tuples, metric_results)

        # Chain them together
        metrics_chained = list(itertools.chain.from_iterable(metric_tuples))
        LOGGER.info("[%s] Calculated all metrics", self.name)
        # Sort the tuples in descending order, so that the key with the highest score
        # is on the first position of the list
        sorted_tuples = multisort(metrics_chained)
        LOGGER.info("[%s] Sorted all metrics", self.name)
        # Retrieve the keys from the tuples
        ordered_key_metric_pairs = [list_unzip(sublist) for sublist in sorted_tuples]
        return ordered_key_metric_pairs

    def update_ordering(self) -> bool:
        """Calculates the ordering for Machine Learning based AL methods
        
        Returns
        -------
        bool
            True if updating succeeded
        """
        
        try:
            self.retrain()
            orderings = self.calculate_ordering()
        except (NotFittedError, IndexError, 
                ValueError, NoLabeledDataException, 
                NoVectorsException) as ex:
            self._uses_fallback = True
            LOGGER.error("[%s] Falling back to model %s, because of: %s",
                                 self.name, self.fallback.name, ex, exc_info=ex)
            self.fallback.update_ordering()
            self._set_ordering([])
            return False
        else:
            self._uses_fallback = False
            # Update ordering for sublearners
            for (ord, metrics) , learner in zip(orderings, self.learners):
                assert isinstance(learner, FixedOrdering)
                learner.enter_ordering(ord, metrics)
            self._has_ordering = True
            return True


    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]:
        """Internal functions that selects the next active learner for the next query

        Returns:
            ActiveLearner[IT, KT, DT, VT, RT, LT]: One of the learners from the ensemble
        """        
        idxs = np.arange(len(self.learners))
        al_idx: int = self._rng.choice(idxs, size=1, p=self.probabilities)[0]
        learner = self.learners[al_idx]
        return learner
    
    def __next__(self) -> IT:
        if not self._uses_fallback:
            result = super().__next__()
            while result.identifier in self.sampled:
                result = super().__next__()
            self.sampled.add(result.identifier)
            return result
        result = next(self.fallback)
        return result



class LabelProbEnsemble(ProbabilityBasedEnsemble[IT, KT, DT, RT, LT], 
                        Generic[IT, KT, DT, RT, LT]):

    def __init__(self,
                 classifier: AbstractClassifier[KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]],
                 strategy: Callable[..., AbstractSelectionCriterion],
                 batch_size = 200,
                 rng: Any = None,
                 identifier: Optional[str] = None, 
                 fallback = RandomSampling[IT, KT, DT, npt.NDArray[Any], RT, LT].builder(), *_, **__) -> None:
        super().__init__(classifier, [], [], 
                         batch_size=batch_size, 
                         identifier=identifier,
                         rng=rng, 
                         fallback=fallback)
        self._strategy_builder = strategy

    def __call__(self, environment: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT]) -> LabelProbEnsemble[IT, KT, DT, RT, LT]:
        super().__call__(environment)
        labelset = list(self.env.labels.labelset)
        self.label_dict = {label: idx for idx, label in enumerate(labelset)}
        label_columns = list(map(self.classifier.get_label_column_index, labelset))
        self.strategies = list(map(self._strategy_builder, label_columns))
        self.probabilities = get_probabilities(None, self.strategies)
        zipped = zip(labelset, self.strategies)
        self.learners = [
            FixedOrdering[IT, KT,DT, npt.NDArray[Any], RT, LT](env, identifier=strategy.name, label=label) 
            for label, strategy in zipped
        ]
        return self

   

class LabelMinProbEnsemble(LabelProbEnsemble[IT, KT, DT, RT, LT], 
                             Generic[IT, KT, DT, RT, LT]):
    
    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]:
        labelcounts = [(self.env.labels.document_count(label), label)
                       for label in self.env.labels.labelset]
        min_label = min(labelcounts)[1]
        al_idx = self.label_dict[min_label]
        learner = self.learners[al_idx]
        return learner
    



from typing import Dict, Generic, Optional, Sequence, Tuple

from allib.environment.base import AbstractEnvironment

from ..typehints import DT, IT, KT, LT, RT, VT
from .random import PoolBasedAL


class FixedOrdering(
    PoolBasedAL[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):

    _name = "FixedOrdering"

    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        *_,
        identifier: Optional[str] = None,
        label: Optional[LT] = None,
        **__,
    ) -> None:
        super().__init__(env, identifier=identifier)
        self.metrics: Dict[KT, float] = dict()
        self.label = label

    @property
    def name(self) -> Tuple[str, Optional[LT]]:
        if self.identifier is not None:
            return f"{self.identifier}", self.label
        return f"{self._name}", self.label

    def enter_ordering(
        self, ordering: Sequence[KT], metrics: Optional[Sequence[float]] = None
    ):
        self._set_ordering(ordering)
        if metrics is not None:
            self.metrics = {o: m for (o, m) in zip(ordering, metrics)}



import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Mapping, Sequence
from uuid import UUID

import pandas as pd

from allib.analysis.analysis import loss_er

from ..utils.func import hn

from .statistics import TarDatasetStats
from .tarplotter import TarExperimentPlotter


@dataclass
class BenchmarkResult:
    dataset: str
    run_id: UUID
    method: str

    stop_wss: float
    stop_recall: float
    stop_loss_er: float
    stop_found: int
    stop_effort: int
    stop_prop_effort: float
    stop_relative_error: float
    
    dataset_stats: TarDatasetStats


def read_datasets(root_path: Path) -> Mapping[str, Mapping[UUID, TarExperimentPlotter]]:
    def read_ds(ds_path: Path):
        for file in ds_path.iterdir():
            if file.suffix == ".pkl":
                with file.open("rb") as fh:
                    plotter = pickle.load(fh)
                run_id = UUID(file.stem.split("_")[1])
                yield (run_id, plotter)

    dss = {ds_path.stem: dict(read_ds(ds_path)) for ds_path in root_path.iterdir()}
    return dss


def extract_results(
    dataset: str, run_id: UUID, plotter: TarExperimentPlotter, crit_name: str,
    target_recall: float = 0.95
) -> BenchmarkResult:
    stop_it = plotter._it_at_stop(crit_name) or plotter.it
    rs = plotter.recall_stats[stop_it]
    ds = plotter.dataset_stats[stop_it]
    l_er = loss_er(rs.pos_docs_found, rs.effort, ds.pos_count, ds.size)
    re = abs(target_recall - rs.recall) / target_recall 
    return BenchmarkResult(
        dataset,
        run_id,
        crit_name,
        rs.wss,
        rs.recall,
        l_er,
        rs.pos_docs_found,
        rs.effort,
        rs.proportional_effort,
        re,
        ds
    )


def extract_information(
    run_dict: Mapping[str, Mapping[UUID, TarExperimentPlotter]]
) -> Sequence[BenchmarkResult]:
    records = [
        extract_results(dataset, run_id, plotter, crit_name)
        for (dataset, runs) in run_dict.items()
        for (run_id, plotter) in runs.items()
        for crit_name in plotter.criterion_names
    ]
    return records

def results_to_pandas(results: Sequence[BenchmarkResult]) -> pd.DataFrame:
    df = pd.DataFrame(results)
    df = pd.concat((df, df.dataset_stats.apply(pd.Series)), axis=1)
    df = df.drop("dataset_stats", axis=1)
    return df



import logging
from pathlib import Path
from typing import Any, Callable, Mapping, TypeVar, Union
from uuid import UUID

import numpy as np
import numpy.typing as npt
import pandas as pd
from instancelib import TextInstance
from instancelib.ingest.spreadsheet import read_csv_dataset
from instancelib.typehints.typevars import KT, LT
from ..typehints.typevars import IT

from ..analysis.experiments import ExperimentIterator
from ..analysis.initialization import Initializer, SeparateInitializer
from ..analysis.simulation import TarSimulator, initialize_tar_simulation
from ..analysis.tarplotter import ModelStatsTar, TarExperimentPlotter
from ..environment import AbstractEnvironment
from ..environment.memory import MemoryEnvironment
from ..estimation.base import AbstractEstimator
from ..module.factory import MainFactory
from ..stopcriterion.base import AbstractStopCriterion
from ..utils.func import list_unzip3

POS = "Relevant"
NEG = "Irrelevant"
LOGGER = logging.getLogger(__name__)


def binary_mapper(value: Any) -> str:
    return POS if value == 1 else NEG


DLT = TypeVar("DLT")
LT = TypeVar("LT")


def read_review_dataset(
    path: Path,
) -> AbstractEnvironment[
    TextInstance[Union[int, UUID], npt.NDArray[Any]],
    Union[int, UUID],
    str,
    npt.NDArray[Any],
    str,
    str,
]:
    """Convert a CSV file with a Systematic Review dataset to a MemoryEnvironment.

    Parameters
    ----------
    path : Path
        The path to the CSV file

    Returns
    -------
    MemoryEnvironment[int, str, npt.NDArray[Any], str]
        A MemoryEnvironment. The labels that
    """
    df = pd.read_csv(path)
    if "label_included" in df.columns:
        env = read_csv_dataset(
            path,
            data_cols=["title", "abstract"],
            label_cols=["label_included"],
            label_mapper=binary_mapper,
        )
    else:
        env = read_csv_dataset(
            path,
            data_cols=["title", "abstract"],
            label_cols=["included"],
            label_mapper=binary_mapper,
        )
    al_env = MemoryEnvironment.from_instancelib_simulation(env)
    return al_env


def benchmark(
    env: AbstractEnvironment[IT, KT, Any, Any, Any, str],
    output_path: Path,
    output_pdf_path: Path,
    al_config: Mapping[str, Any],
    fe_config: Mapping[str, Any],
    initializer_builder: Callable[..., Initializer[IT, KT, str]],
    estimators: Mapping[str, AbstractEstimator[IT, KT, Any, Any, Any, str]],
    stopcriteria: Mapping[str, AbstractStopCriterion[str]],
    pos_label: str,
    neg_label: str,
    batch_size: int = 10,
    stop_interval: Union[int, Mapping[str, int]] = 10,
    estimation_interval: Union[int, Mapping[str, int]] = 10,
) -> TarExperimentPlotter[str]:
    factory = MainFactory()
    initializer = initializer_builder(pos_label=pos_label, neg_label=neg_label)
    al, _ = initialize_tar_simulation(
        factory, al_config, fe_config, initializer, env, pos_label, neg_label
    )
    exp = ExperimentIterator(
        al,
        pos_label,
        neg_label,
        stopcriteria,
        estimators,
        batch_size,
        stop_interval,
        estimation_interval,
    )
    plotter = ModelStatsTar(POS, NEG)
    simulator = TarSimulator(
        exp, plotter, output_path=output_path, output_pdf_path=output_pdf_path
    )
    try:
        simulator.simulate()
    except Exception as e:
        LOGGER.error("Exited with %s", e)
        pass
    return plotter



from enum import Enum
from typing import Dict

from ..factory import AbstractBuilder, ObjectFactory
from ..module.component import Component

from .base import BaseBalancer, IdentityBalancer
from .catalog import BalancerCatalog as BL
from .double import DoubleBalancer
from .randomoversampling import RandomOverSampler
from .undersample import UndersampleBalancer


class BalancerBuilder(AbstractBuilder):
    def __call__(self, # type: ignore
            type: BL.Type, 
            config: Dict, **kwargs) -> BaseBalancer:
        return self._factory.create(type, **config)

class BalancerFactory(ObjectFactory):
    def __init__(self) -> None:
        super().__init__()
        self.register_builder(
            Component.BALANCER, BalancerBuilder())
        self.register_constructor(
            BL.Type.IDENTITY, IdentityBalancer)
        self.register_constructor(
            BL.Type.UNDERSAMPLING, UndersampleBalancer)
        self.register_constructor(
            BL.Type.RANDOM_OVER_SAMPLING, RandomOverSampler)
        self.register_constructor(
            BL.Type.DOUBLE, DoubleBalancer)



from enum import Enum

class BalancerCatalog:
    class Type(Enum):
        IDENTITY = "Identity"
        RANDOM_OVER_SAMPLING = "RandomOverSampling"
        UNDERSAMPLING = "UnderSampling"
        DOUBLE = "DoubleBalancer"